import pytest
from pandas import DataFrame

from bender.importers import DataImporters
from bender.split_strategies import SplitStrategies

pytestmark = pytest.mark.asyncio


async def test_start_ratio(date_df: DataFrame) -> None:

    ratio = 0.8

    for offset in range(0, 10):
        train, test = await (
            DataImporters.literal(date_df)
            .split(SplitStrategies.uniform_ratio('bool_classification', ratio, start_offset=offset / 10))
            .run()
        )



import pytest
from pandas.core.frame import DataFrame

from bender.importers import DataImporters
from bender.transformations import Transformations

pytestmark = pytest.mark.asyncio


async def test_cached_importer(date_df: DataFrame) -> None:
    cache_path = 'test-cache'
    processing = [Transformations.log_normal_shift('y_values', 'y_log_values')]
    original_data = await DataImporters.literal(date_df).cached(cache_path).process(processing).run()
    cached_data = await DataImporters.literal(DataFrame({})).cached(cache_path).process(processing).run()

    cached_data.set_index(original_data.index)
    assert len(cached_data.columns) == len(original_data.columns)



# type: ignore[misc]
from __future__ import annotations

import asyncio
import logging
from datetime import datetime, timedelta
from typing import Any, Callable, Optional

from pandas.core.frame import DataFrame
from pandas.core.series import Series

from bender.data_exporter.disk import DiskDataExporter
from bender.data_exporter.interface import DataExporter
from bender.data_exporter.pipeline import Extractable
from bender.data_importer.importer import AppendImporter, CachedImporter, DataImportable, DataImporter, JoinedImporter
from bender.evaluator.factory_method import Evaluable
from bender.evaluator.interface import Evaluator
from bender.explorer.factory import Explorable
from bender.explorer.interface import Explorer
from bender.metric.factory import Metricable
from bender.metric.interface import Metric
from bender.model_exporter.factory import ModelExportable
from bender.model_exporter.interface import ModelExporter
from bender.model_loader.model_loader import ModelLoadable, ModelLoader
from bender.model_trainer.interface import ModelTrainer, Trainable
from bender.pipeline.interface import RunnablePipeline
from bender.prediction_extractor import Predictable, PredictionExtractor, PredictionOutput, ProbabilisticPredictable
from bender.split_strategies import SplitStrategies
from bender.split_strategy.split_strategy import Splitable, SplitStrategy, TrainingDataSet
from bender.trained_model.interface import TrainedModel, TrainedProbabilisticModel
from bender.transformation.transformation import Processable, Transformation

logger = logging.getLogger(__name__)


class ExplorePipeline(RunnablePipeline[DataFrame], Splitable):

    pipeline: RunnablePipeline[DataFrame]
    explorers: list[Explorer]

    def __init__(self, pipeline: RunnablePipeline[DataFrame], explorers: list[Explorer]) -> None:
        self.pipeline = pipeline
        self.explorers = explorers

    async def run(self) -> DataFrame:
        df = await self.pipeline.run()
        for explorer in self.explorers:
            await explorer.explore(df)
        return df

    def split(self, split_strategy: SplitStrategy) -> SplitedData:
        return SplitedData(self, split_strategy)


class LoadedDataAndModel(
    Processable,
    Predictable,
    ProbabilisticPredictable,
    RunnablePipeline[tuple[TrainedModel, DataFrame]],
):

    data_loader: LoadedData
    model_loader: LoadedModel

    def __init__(self, data_loader: LoadedData, model_loader: LoadedModel) -> None:
        self.data_loader = data_loader
        self.model_loader = model_loader

    def process(self, transformations: list[Transformation]) -> LoadedDataAndModel:
        # Some processing of the data
        return LoadedDataAndModel(self.data_loader.process(transformations), self.model_loader)

    def predict(self, on: Optional[Callable[[DataFrame], Series]] = None) -> PredictionPipeline:
        return PredictionPipeline(self, predict_on=on)

    def predict_proba(self, on: Optional[Callable[[DataFrame], Series]] = None) -> ProbalisticPredictionPipeline:
        return ProbalisticPredictionPipeline(self, predict_on=on)

    async def run(self) -> tuple[TrainedModel, DataFrame]:
        return await asyncio.gather(self.model_loader.run(), self.data_loader.run())


class Extractor(RunnablePipeline[DataFrame]):

    pipeline: RunnablePipeline[tuple[TrainedModel, DataFrame, Series]]
    prediction_feature: str
    features: list[str]
    exporter: DataExporter
    transformations: list[Transformation]

    def __init__(
        self,
        pipeline: RunnablePipeline[tuple[TrainedModel, DataFrame, Series]],
        prediction_feature: str,
        features: list[str],
        exporter: DataExporter,
        transformations: list[Transformation],
    ) -> None:
        self.pipeline = pipeline
        self.prediction_feature = prediction_feature
        self.features = features
        self.exporter = exporter
        self.transformations = transformations

    async def run(self) -> DataFrame:
        model, data, predictions = await self.pipeline.run()

        extraction = DataFrame()
        if self.features:
            extraction = data[self.features]
        extraction[self.prediction_feature] = predictions
        for transformation in self.transformations:
            extraction = await transformation.transform(extraction)
        await self.exporter.export(extraction)
        return data


class ExtractorProbability(RunnablePipeline[DataFrame]):

    pipeline: RunnablePipeline[tuple[TrainedModel, DataFrame, Series]]
    prediction_features: dict[Any, str]
    features: list[str]
    exporter: DataExporter
    transformations: list[Transformation]

    def __init__(
        self,
        pipeline: RunnablePipeline[tuple[TrainedModel, DataFrame, Series]],
        prediction_features: dict[Any, str],
        features: list[str],
        exporter: DataExporter,
        transformations: list[Transformation],
    ) -> None:
        self.pipeline = pipeline
        self.prediction_features = prediction_features
        self.features = features
        self.exporter = exporter
        self.transformations = transformations

    async def run(self) -> DataFrame:
        model, data, predictions = await self.pipeline.run()

        extraction = DataFrame()
        if self.features:
            extraction = data[self.features].copy()
        for classification, output_name in self.prediction_features.items():
            extraction[output_name] = predictions[classification].copy()
        for transformation in self.transformations:
            extraction = await transformation.transform(extraction)
        await self.exporter.export(extraction)
        return data


class PredictionPipeline(RunnablePipeline[tuple[TrainedModel, DataFrame, Series]], Extractable):

    data_and_model: LoadedDataAndModel
    predict_on: Optional[Callable[[DataFrame], Series]]

    def __init__(self, data_and_model: LoadedDataAndModel, predict_on: Optional[Callable[[DataFrame], Series]]) -> None:
        self.data_and_model = data_and_model
        self.predict_on = predict_on

    async def run(self) -> tuple[TrainedModel, DataFrame, Series]:
        model, processed_data = await self.data_and_model.run()

        if self.predict_on:
            predict_on_filter = self.predict_on(processed_data)
            predict_on_data = processed_data[predict_on_filter]
        else:
            predict_on_data = processed_data
        result = model.predict(predict_on_data)
        return (model, processed_data, result)

    def extract(
        self,
        prediction_as: str,
        metadata: Optional[list[str]] = None,
        exporter: DataExporter = DiskDataExporter('predictions.csv'),
        transforations: Optional[list[Transformation]] = None,
    ) -> Extractor:
        return Extractor(
            self,
            prediction_as,
            [] if metadata is None else metadata,
            exporter,
            [] if transforations is None else transforations,
        )


class CrossValidate(RunnablePipeline[float]):

    target: str
    k_fold: int
    pipeline: RunnablePipeline[DataFrame]

    def __init__(
        self,
        pipeline: RunnablePipeline[DataFrame],
        target: str,
        k_fold: int,
        validation: Callable[[SplitedData], RunnablePipeline[float]],
    ) -> None:
        self.k_fold = k_fold
        self.target = target
        self.pipeline = pipeline
        self.validation = validation

    async def run(self) -> float:
        train_ratio = (self.k_fold - 1) / self.k_fold
        offset_ratio = 1 / self.k_fold
        score_sum = float(0)
        numb_of_scores = 0
        for i in range(0, self.k_fold):
            data_pipe = SplitedData(
                self.pipeline, SplitStrategies.uniform_ratio(self.target, train_ratio, offset_ratio * i)
            )
            pipe = self.validation(data_pipe)
            score = await pipe.run()

            if not score or not isinstance(score, (float, int)):
                logger.info(f'K-fold corss validation returned invalid metric: {score}')
                continue

            score_sum += score
            numb_of_scores += 1

        if numb_of_scores > 0:
            return score_sum / numb_of_scores
        else:
            return float('nan')


class ProbalisticPredictionPipeline(RunnablePipeline[tuple[TrainedModel, DataFrame, Series]]):

    data_and_model: RunnablePipeline[tuple[TrainedProbabilisticModel, DataFrame]]
    predict_on: Optional[Callable[[DataFrame], Series]]

    def __init__(
        self,
        data_and_model: RunnablePipeline[tuple[TrainedProbabilisticModel, DataFrame]],
        predict_on: Optional[Callable[[DataFrame], Series]],
    ) -> None:
        self.data_and_model = data_and_model
        self.predict_on = predict_on

    async def run(self) -> tuple[TrainedModel, DataFrame, Series]:
        model, processed_data = await self.data_and_model.run()

        if self.predict_on:
            predict_on_filter = self.predict_on(processed_data)
            predict_on_data = processed_data[predict_on_filter]
        else:
            predict_on_data = processed_data

        return (model, processed_data, model.predict_proba(predict_on_data))

    def extract(
        self,
        prediction_as: dict[Any, str],
        metadata: Optional[list[str]] = None,
        exporter: DataExporter = DiskDataExporter('predictions.csv'),
        transforations: Optional[list[Transformation]] = None,
    ) -> ExtractorProbability:
        return ExtractorProbability(
            self,
            prediction_as,
            [] if metadata is None else metadata,
            exporter,
            [] if transforations is None else transforations,
        )


class ExtractFromPredictionPipeline(RunnablePipeline[None]):

    data_and_model: LoadedDataAndModel
    predict_on: Optional[Callable[[DataFrame], Series]]
    extractors: list[PredictionExtractor]

    def __init__(
        self,
        data_and_model: LoadedDataAndModel,
        predict_on: Callable[[DataFrame], Series],
        extractors: list[PredictionExtractor],
    ) -> None:
        self.data_and_model = data_and_model
        self.predict_on = predict_on
        self.extractors = extractors

    async def run(self) -> None:
        model, processed_data = await self.data_and_model.run()
        if self.predict_on:
            predict_on_filter = self.predict_on(processed_data)
            predict_on_data = processed_data[predict_on_filter]
        else:
            predict_on_data = processed_data

        for extractor in self.extractors:
            output_feature, output_type = extractor.output

            output_df = DataFrame()
            if output_type == PredictionOutput.CLASSIFICATION:
                output_df[output_feature] = model.predict(predict_on_data)
            elif output_type == PredictionOutput.PROBABILITY:
                output_df[output_feature] = model.predict_proba(predict_on_data)
            else:
                raise NotImplementedError()

            for needed_feature in extractor.needed_features:
                output_df[needed_feature] = predict_on_data[needed_feature]

            await extractor.extract(output_df)


class LoadedModel(DataImportable[LoadedDataAndModel], RunnablePipeline[TrainedModel], ModelLoader):

    model_loader: ModelLoader

    def __init__(self, model_loader: ModelLoader) -> None:
        self.model_loader = model_loader

    def import_data(self, importer: DataImporter) -> LoadedDataAndModel:
        return LoadedDataAndModel(importer, self)

    async def load_model(self) -> TrainedModel:
        return await self.run()

    async def run(self) -> TrainedModel:
        return await self.model_loader.load_model()


class LoadedData(
    Processable,
    ModelLoadable[LoadedDataAndModel],
    RunnablePipeline[DataFrame],
    Splitable,
    DataImporter,
    Explorable[ExplorePipeline],
):

    importer: DataImporter
    transformations: list[Transformation]

    def __init__(self, importer: DataImporter, transformations: list[Transformation]) -> None:
        self.importer = importer
        self.transformations = transformations
        assert isinstance(importer, DataImporter)

    def process(self, transformations: list[Transformation]) -> LoadedData:
        # Some processing of the data
        return LoadedData(self.importer, self.transformations + transformations)

    def load_model(self, model_loader: ModelLoader) -> LoadedDataAndModel:
        return LoadedDataAndModel(self, LoadedModel(model_loader))

    async def import_data(self) -> DataFrame:
        return await self.run()

    async def run(self) -> DataFrame:
        logger.info('Fetching data')
        df = await self.importer.import_data()
        logger.info('Fetched data')

        for transformation in self.transformations:
            logger.info(f'Applying transformation {type(transformation)}')
            df = await transformation.transform(df)
            if df.empty:
                logger.error(f'DataFrame ended up as empty after a {type(transformation)}. Exiting early')
                raise Exception('Transformations returned an empty data set')
        return df

    def split(self, split_strategy: SplitStrategy) -> SplitedData:
        return SplitedData(self, split_strategy=split_strategy)

    def cross_validate(
        self, target: str, k_fold: int, validation: Callable[[SplitedData], RunnablePipeline]
    ) -> CrossValidate:
        return CrossValidate(self, target, k_fold, validation)

    def join_import(self, importer: DataImporter, join_key: str) -> LoadedData:
        return LoadedData(
            JoinedImporter(first_import=self.importer, second_import=importer, join_key=join_key), self.transformations
        )

    def cached(
        self, path: str, from_now: Optional[timedelta] = None, timestamp: Optional[datetime] = None
    ) -> LoadedData:
        importer: CachedImporter
        if timestamp:
            importer = CachedImporter(self.importer, path, timestamp)
        elif from_now:
            importer = CachedImporter(self.importer, path, datetime.now() + from_now)
        else:
            importer = CachedImporter(self.importer, path, datetime.now() + timedelta(days=1))
        return LoadedData(importer, self.transformations)

    def append(self, importer: LoadedData, ignore_index: bool = True) -> LoadedData:
        """Append two differnet data importers.
        This can be usefull when you have different types of data, but with the same features

        Args:
            importer (DataImporter): The data to append

        Returns:
            DataImporter: A Importer that appends the multiple importers
        """
        return LoadedData(AppendImporter(self.importer, importer, ignore_index=ignore_index), self.transformations)

    def explore(self, explorers: list[Explorer]) -> ExplorePipeline:
        return ExplorePipeline(self, explorers)


class SplitedData(RunnablePipeline[tuple[DataFrame, DataFrame]], Trainable):

    data_loader: RunnablePipeline[DataFrame]
    split_strategy: SplitStrategy

    def __init__(self, data_loader: RunnablePipeline[DataFrame], split_strategy: SplitStrategy) -> None:
        self.data_loader = data_loader
        self.split_strategy = split_strategy

    async def run(self) -> tuple[DataFrame, DataFrame]:
        raw_data = await self.data_loader.run()
        return await self.split_strategy.split(raw_data)

    def train(self, model: ModelTrainer, input_features: list[str], target_feature: str) -> TrainingPipeline:
        return TrainingPipeline(self, model, input_features, target_feature)


class ValidationLoss(RunnablePipeline[float]):

    pipeline: RunnablePipeline[tuple[TrainedModel, TrainingDataSet]]

    def __init__(self, pipeline: RunnablePipeline[tuple[TrainedModel, TrainingDataSet]]) -> None:
        self.pipeline = pipeline

    async def run(self) -> float:
        model, data_set = await self.pipeline.run()
        return 0


class LossValidatable:
    def loss(self) -> ValidationLoss:
        raise NotImplementedError()


class PipelineMetric(RunnablePipeline[float]):

    metric: Metric
    pipeline: RunnablePipeline[tuple[TrainedModel, TrainingDataSet]]

    def __init__(self, pipeline: RunnablePipeline[tuple[TrainedModel, TrainingDataSet]], metric: Metric) -> None:
        self.metric = metric
        self.pipeline = pipeline

    async def run(self) -> float:
        model, data_set = await self.pipeline.run()
        return await self.metric.metric(model, data_set)


class ExportTrainedModel(
    RunnablePipeline[tuple[TrainedModel, TrainingDataSet]], LossValidatable, Metricable[PipelineMetric]
):

    pipeline: RunnablePipeline[tuple[TrainedModel, TrainingDataSet]]
    exporter: ModelExporter

    def __init__(
        self, pipeline: RunnablePipeline[tuple[TrainedModel, TrainingDataSet]], exporter: ModelExporter
    ) -> None:
        self.pipeline = pipeline
        self.exporter = exporter

    async def run(self) -> tuple[TrainedModel, TrainingDataSet]:
        model, data_set = await self.pipeline.run()
        await self.exporter.export(model)
        return model, data_set

    def metric(self, metric: Metric) -> PipelineMetric:
        return PipelineMetric(self, metric)


class TrainAndEvaluatePipeline(
    RunnablePipeline[tuple[TrainedModel, TrainingDataSet]],
    ModelExportable[ExportTrainedModel],
    LossValidatable,
    Metricable[PipelineMetric],
):

    pipeline: RunnablePipeline[tuple[TrainedModel, TrainingDataSet]]
    evaluators: list[Evaluator]

    def __init__(
        self, pipeline: RunnablePipeline[tuple[TrainedModel, TrainingDataSet]], evaluators: list[Evaluator]
    ) -> None:
        self.pipeline = pipeline
        self.evaluators = evaluators

    async def run(self) -> tuple[TrainedModel, TrainingDataSet]:
        model, train_set = await self.pipeline.run()
        for evaluator in self.evaluators:
            logger.info(f'Evaluating with evaloator: {type(evaluator)}')
            await evaluator.evaluate(model, train_set)
        return model, train_set

    def export_model(self, exporter: ModelExporter) -> ExportTrainedModel:
        return ExportTrainedModel(self, exporter)

    def metric(self, metric: Metric) -> PipelineMetric:
        return PipelineMetric(self, metric)


class TrainingPipeline(
    RunnablePipeline[tuple[TrainedModel, TrainingDataSet]],
    Evaluable[TrainAndEvaluatePipeline],
    ModelExportable[ExportTrainedModel],
    LossValidatable,
    Metricable[PipelineMetric],
):

    data_loader: SplitedData
    model_trainer: ModelTrainer
    input_features: list[str]
    target_feature: str

    def __init__(
        self, data_loader: SplitedData, model_trainer: ModelTrainer, input_features: list[str], target_feature: str
    ) -> None:
        self.data_loader = data_loader
        self.model_trainer = model_trainer
        self.input_features = input_features
        self.target_feature = target_feature

    async def _training_set(self) -> TrainingDataSet:
        test_data, validation_data = await self.data_loader.run()
        return TrainingDataSet(self.input_features, self.target_feature, test_data, validation_data)

    async def run(self) -> tuple[TrainedModel, TrainingDataSet]:
        train_set = await self._training_set()
        model = await self.model_trainer.train(train_set)
        return model, train_set

    def evaluate(self, evaluators: list[Evaluator]) -> TrainAndEvaluatePipeline:
        return TrainAndEvaluatePipeline(self, evaluators)

    def export_model(self, exporter: ModelExporter) -> ExportTrainedModel:
        return ExportTrainedModel(self, exporter)

    def metric(self, metric: Metric) -> PipelineMetric:
        return PipelineMetric(self, metric)



from __future__ import annotations

import logging
from typing import Any, Callable, Generic, Optional, TypeVar

import numpy as np
import pandas as pd
from numpy import datetime64, log1p, log2
from pandas import DataFrame
from pandas.core.groupby.generic import SeriesGroupBy
from pandas.core.series import Series
from sklearn.neighbors import BallTree

logger = logging.getLogger(__name__)


class Transformation:
    async def transform(self, df: DataFrame) -> DataFrame:
        raise NotImplementedError()

    def only_if(self, should_run: bool) -> Transformation:
        return OptionalTransformation(self, should_run=should_run)


ProcessResultType = TypeVar('ProcessResultType')


class Processable(Generic[ProcessResultType]):
    def process(self, transformations: list[Transformation]) -> ProcessResultType:
        raise NotImplementedError()


class OptionalTransformation(Transformation):

    should_run: bool
    transformation: Transformation

    def __init__(self, transformation: Transformation, should_run: bool) -> None:
        self.transformation = transformation
        self.should_run = should_run

    async def transform(self, df: DataFrame) -> DataFrame:
        if self.should_run:
            return await self.transformation.transform(df)
        return df


class RemoveDuplicateColumns(Transformation):
    async def transform(self, df: DataFrame) -> DataFrame:
        s = df.columns.to_series()
        df.columns = s + s.groupby(s).cumcount().astype(str).replace({'0': ''})
        return df


class DescribeData(Transformation):

    features: Optional[list[str]]

    def __init__(self, features: Optional[list[str]] = None) -> None:
        self.features = features

    async def transform(self, df: DataFrame) -> DataFrame:
        display_features = list(df.columns)
        if self.features:
            display_features = self.features
        logger.info(df[display_features].describe())
        return df


class LogDataFeatures(Transformation):
    async def transform(self, df: DataFrame) -> DataFrame:
        logger.info(df.columns)
        return df


class LogDataHead(Transformation):

    features: Optional[list[str]]

    def __init__(self, features: Optional[list[str]] = None) -> None:
        self.features = features

    async def transform(self, df: DataFrame) -> DataFrame:
        display_features = list(df.columns)
        if self.features:
            display_features = self.features
        logger.info(df[display_features[0]].iloc[0])
        logger.info(df[display_features].head())
        return df


class LogDataInfo(Transformation):
    async def transform(self, df: DataFrame) -> DataFrame:
        logger.info(df.info())
        return df


class Filter(Transformation):
    def __init__(self, lambda_function: Callable[[DataFrame], Series]) -> None:
        self.lambda_function = lambda_function

    async def transform(self, df: DataFrame) -> DataFrame:
        return df[self.lambda_function(df)].copy()


class SetIndex(Transformation):

    column: str

    def __init__(self, column: str) -> None:
        self.column = column

    async def transform(self, df: DataFrame) -> DataFrame:
        df.set_index(self.column, inplace=True)
        return df


class LogNormalDistributionShift(Transformation):

    input_feature: str
    output_feature: str
    input_has_zeros: bool

    def __init__(self, input_feature: str, output_feature: Optional[str] = None, input_has_zeros: bool = True) -> None:
        self.input_feature = input_feature
        self.input_has_zeros = input_has_zeros
        if output_feature:
            self.output_feature = output_feature
        else:
            self.output_feature = f'{input_feature}_log'

    async def transform(self, df: DataFrame) -> DataFrame:
        if self.input_has_zeros:
            df[self.output_feature] = log1p(df[self.input_feature])
        else:
            df[self.output_feature] = log2(df[self.input_feature]) / log2(np.e)
        return df


class DateComponent(Transformation):
    # ValueError: Tz-aware datetime.datetime cannot be converted to datetime64 unless utc=True

    input_feature: str
    output_feature: str
    component: str

    def __init__(self, component: str, input_feature: str, output_feature: Optional[str] = None) -> None:

        self.component = component
        self.input_feature = input_feature
        if output_feature:
            self.output_feature = output_feature
        else:
            self.output_feature = f'{self.input_feature}_{component}'

    async def transform(self, df: DataFrame) -> DataFrame:
        if df[self.input_feature].dtype != datetime64 or pd.Timestamp:
            df[self.input_feature] = pd.to_datetime(df[self.input_feature], infer_datetime_format=True, utc=True)
        df[self.output_feature] = getattr(df[self.input_feature].dt, self.component)
        return df


class BinaryTransform(Transformation):

    output_feature: str

    def __init__(self, output_feature: str, lambda_function: Callable[[DataFrame], Series]) -> None:
        self.output_feature = output_feature
        self.lambda_function = lambda_function

    async def transform(self, df: DataFrame) -> DataFrame:
        df[self.output_feature] = self.lambda_function(df)
        if df[self.output_feature].dtype == bool:
            df[self.output_feature] = df[self.output_feature].astype(int)
        return df


class SelectFeatures(Transformation):

    features: list[str]

    def __init__(self, features: list[str]) -> None:
        self.features = features

    async def transform(self, df: DataFrame) -> DataFrame:
        return df[self.features]


class DiscardFeatures(Transformation):

    features: list[str]

    def __init__(self, features: list[str]) -> None:
        self.features = features

    async def transform(self, df: DataFrame) -> DataFrame:
        all_features = list(df.columns)
        for remove_feature in self.features:
            if remove_feature in all_features:
                all_features.remove(remove_feature)
        return df[all_features]


class UnpackList(Transformation):

    input_feature: str
    index: int
    output_feature: str

    def __init__(self, input_feature: str, index: int, output_feature: str) -> None:
        self.input_feature = input_feature
        self.index = index
        self.output_feature = output_feature

    async def transform(self, df: DataFrame) -> DataFrame:
        df[self.output_feature] = df[self.input_feature].apply(pd.Series).loc[:, self.index]
        return df


class UnpackPolicy:
    def unpack(self, column: Series, key: str) -> Series:
        raise NotImplementedError()

    @staticmethod
    def min_number() -> UnpackPolicy:
        return UnpackNumber(lambda df: df.min())

    @staticmethod
    def max_number() -> UnpackPolicy:
        return UnpackNumber(lambda df: df.max())

    @staticmethod
    def median_number() -> UnpackPolicy:
        return UnpackNumber(lambda df: df.median())

    @staticmethod
    def mean_number() -> UnpackPolicy:
        return UnpackNumber(lambda df: df.mean())

    @staticmethod
    def first_string() -> UnpackPolicy:
        return UnpackString()


class UnpackNumber(UnpackPolicy):
    def __init__(self, metric: Callable[[SeriesGroupBy], Series]) -> None:
        self.metric = metric

    def unpack(self, column: Series, key: str) -> Series:
        bracket = '}'
        regex_str = rf'"{key}"[\s:]+(\d+)[{bracket},]'
        grouped = column.astype(str).str.extractall(regex_str).reset_index(level='match', drop=True).reset_index()
        grouped.columns = ['index', 'value']
        # Is needed to set type to float, as it can otherwise lead to bugs
        # Like incorrect mean values
        grouped['value'] = grouped['value'].astype(float)
        grouped = grouped.groupby(by='index')

        return self.metric(grouped).astype(float)


class UnpackString(UnpackPolicy):
    def unpack(self, column: Series, key: str) -> Series:
        bracket = '}'
        regex_str = rf'"{key}"[\s:]+"([\w\d\s\\.-]+)["{bracket},]'
        return column.astype(str).str.extract(regex_str).astype(str)


class UnpackJson(Transformation):

    input_feature: str
    key: str
    output_feature: str
    policy: UnpackPolicy

    def __init__(self, input_feature: str, key: str, output_feature: str, policy: UnpackPolicy) -> None:
        self.input_feature = input_feature
        self.key = key
        self.output_feature = output_feature
        self.policy = policy

    async def transform(self, df: DataFrame) -> DataFrame:
        df[self.output_feature] = self.policy.unpack(df[self.input_feature], self.key)
        return df


class NeighbourDistance(Transformation):

    to: Optional[Callable[[DataFrame], Series]]
    latitude_key: str
    longitude_key: str
    number_of_neighbours: int
    output: str

    def __init__(
        self,
        number_of_neighbours: int,
        latitude: str = 'latitude',
        longitude: str = 'longitude',
        to: Optional[Callable[[DataFrame], Series]] = None,
        output: Optional[str] = None,
    ) -> None:
        self.latitude_key = latitude
        self.longitude_key = longitude
        self.number_of_neighbours = number_of_neighbours
        self.to = to
        self.output = 'distance_neighbor' if output is None else output

    async def transform(self, df: DataFrame) -> DataFrame:
        for column in df[[self.latitude_key, self.longitude_key]]:
            if not isinstance(df[column].values[0], float):
                df[column] = df[column].astype(float)
            rad = np.deg2rad(df[column].values)
            df[f'{column}_rad'] = rad

        if self.to:
            to_points = df[self.to(df)]
        else:
            to_points = df

        ball = BallTree(to_points[[f'{self.latitude_key}_rad', f'{self.longitude_key}_rad']].values, metric='haversine')
        distances, _ = ball.query(
            df[[f'{self.latitude_key}_rad', f'{self.longitude_key}_rad']].values, k=self.number_of_neighbours + 1
        )

        # To km
        distances *= 6371

        distances[distances == 0] = np.nan
        df['distance_neighbor'] = np.nanmean(distances, axis=1)

        return df


class DropNaN(Transformation):
    async def transform(self, df: DataFrame) -> DataFrame:
        df.dropna()
        return df


class Relation(Transformation):

    value_key: str
    per_key: str
    output_key: str

    def __init__(self, key: str, per: str, output: str) -> None:
        self.value_key = key
        self.per_key = per
        self.output_key = output

    async def transform(self, df: DataFrame) -> DataFrame:
        mask = df[self.value_key].isna() | df[self.per_key].isna() | df[self.per_key] == 0
        df.loc[mask, self.output_key] = np.nan
        df.loc[~mask, self.output_key] = df[self.value_key] / df[self.per_key]
        return df


class ExponentialShift(Transformation):

    value: str
    output: str

    def __init__(self, value: str, output: str) -> None:
        self.value = value
        self.output = output

    async def transform(self, df: DataFrame) -> DataFrame:
        df[self.output] = np.exp(df[self.value])
        return df


class FillPolicy:
    def fill_value(self, feature: str, df: DataFrame) -> Any:
        raise NotImplementedError()

    @staticmethod
    def constant(value: Any) -> ConstantFillValue:
        return ConstantFillValue(value)

    @staticmethod
    def mean(similar_to: Optional[list[str]] = None) -> MeanFillValue:
        return MeanFillValue(similar_to if similar_to is not None else [])


class ConstantFillValue(FillPolicy):

    value: Any

    def __init__(self, value: Any) -> None:
        self.value = value

    def fill_value(self, feature: str, df: DataFrame) -> Any:
        return self.value


class MeanFillValue(FillPolicy):

    similar_to_features: list[str]

    def __init__(self, similar_to_features: list[str]) -> None:
        self.similar_to_features = similar_to_features

    def fill_value(self, feature: str, df: DataFrame) -> Any:
        if not self.similar_to_features:
            # Find rows most similar to the features
            # and return the mean here
            logger.info('Using based_on_features is not supported yet')
        return df[feature].mean()


class FillMissingValue(Transformation):

    value_key: str
    policy: FillPolicy

    def __init__(self, key: str, policy: FillPolicy) -> None:
        self.value_key = key
        self.policy = policy

    async def transform(self, df: DataFrame) -> DataFrame:
        df.loc[df[self.value_key].isnull(), self.value_key] = self.policy.fill_value(self.value_key, df)
        return df


class CustomCodeTransformation(Transformation):
    def __init__(self, transformation: Callable[[DataFrame], DataFrame]) -> None:
        self.transformation = transformation

    async def transform(self, df: DataFrame) -> DataFrame:
        return self.transformation(df)


class CombineToMean(Transformation):

    output_key: str
    features: list[str]

    def __init__(self, output_key: str, features: list[str]) -> None:
        self.features = features
        self.output_key = output_key

    async def transform(self, df: DataFrame) -> DataFrame:
        df[self.output_key] = df[self.features].mean(axis=1)
        return df


class BinPolicy:
    def bin(column: Series, n_bins: int) -> Series:
        raise NotImplementedError()


class BinFeature(Transformation):

    feature: str
    output: str
    n_bins: int

    def __init__(self, feature: str, output: str, n_bins: int) -> None:
        self.feature = feature
        self.output = output
        self.n_bins = n_bins

    async def transform(self, df: DataFrame) -> DataFrame:
        df[self.output] = 0
        for number in range(self.n_bins):
            percentile = number / self.n_bins
            threshold = df[self.feature].quantile(percentile)
            mask = df[self.feature] > threshold
            df.loc[mask, self.output] = number

        return df


class SplitString(Transformation):

    feature: str
    outputs: list[str]
    seperator: str
    select_number: int

    def __init__(self, feature: str, outputs: list[str], seperator: str, select_number: int) -> None:
        self.feature = feature
        self.outputs = outputs
        self.seperator = seperator
        self.select_number = select_number

    async def transform(self, df: DataFrame) -> DataFrame:
        if df[self.feature].dtype != str:
            df[self.feature] = df[self.feature].astype(str)
        number_of_values = len(self.outputs)
        n = self.select_number + number_of_values - 2  # 0 = 1 value
        result = df[self.feature].str.split(self.seperator, n=n, expand=True)
        df[self.outputs] = result[result.columns[self.select_number - 1 : n + 1]]
        return df


class LogToConsole(Transformation):
    def __init__(self, data_to_log: Callable[[DataFrame], Any]) -> None:
        self.data_to_log = data_to_log

    async def transform(self, df: DataFrame) -> DataFrame:
        data = self.data_to_log(df)
        logger.info(data)
        return df


class ToCategorical(Transformation):

    input: str
    output: str

    def __init__(self, input: str, output: str) -> None:
        self.input = input
        self.output = output

    async def transform(self, df: DataFrame) -> DataFrame:
        df[self.output] = df[self.input].astype('category').cat.codes
        return df



import numpy as np
from sklearn.metrics import log_loss

from bender.metric.interface import Metric
from bender.split_strategy.split_strategy import TrainingDataSet
from bender.trained_model.interface import TrainedModel, TrainedProbabilisticClassificationModel


class LogLoss(Metric):
    async def metric(self, model: TrainedModel, data_set: TrainingDataSet) -> float:
        if isinstance(model, TrainedProbabilisticClassificationModel):
            y_pred = model.predict_proba(data_set.x_validate)
            return float(log_loss(data_set.y_validate, y_pred, labels=model.class_names()))
        else:
            result = model.predict(data_set.x_validate)
            return float(np.log1p(abs(result - data_set.y_validate)).mean())



from datetime import datetime

import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_almost_equal
from pandas.core.frame import DataFrame, Series

from bender.importers import DataImporters
from bender.transformation.schema import SchemaType
from bender.transformation.transformation import UnpackPolicy
from bender.transformations import Transformations

pytestmark = pytest.mark.asyncio


async def test_training_pipeline(date_df: DataFrame) -> None:

    pipeline = DataImporters.literal(date_df).process(
        [
            Transformations.log_normal_shift('y_values', 'y_log', input_has_zeros=False),
            Transformations.neighour_distance(number_of_neighbours=2, latitude='lat', longitude='long'),
            Transformations.date_component('day', 'date', output='day_value'),
            Transformations.date_component('month', 'date', output='month_value'),
            Transformations.date_component('year', 'date', output='year_value'),
        ]
    )
    output_df = await pipeline.run()
    assert_almost_equal(output_df['x_values'].to_numpy(), output_df['y_log'].to_numpy())
    assert np.all(output_df['day_value'] == output_df['expected_day'])
    assert np.all(output_df['month_value'] == output_df['expected_month'])
    assert np.all(output_df['year_value'] == Series([2020] * len(date_df)))


async def run_test_for(policy: UnpackPolicy, name: str, json_df: DataFrame) -> None:
    transformation = Transformations.unpack_json('json_data', key='value', output=f'out_{name}', policy=policy)
    ret_df = await transformation.transform(json_df)
    assert np.all(ret_df[f'out_{name}'] == ret_df[f'value_{name}'])


async def test_unpack_json_median(json_df: DataFrame) -> None:
    await run_test_for(UnpackPolicy.median_number(), 'median', json_df)


async def test_unpack_json_mean(json_df: DataFrame) -> None:
    await run_test_for(UnpackPolicy.mean_number(), 'mean', json_df)


async def test_unpack_json_min(json_df: DataFrame) -> None:
    await run_test_for(UnpackPolicy.min_number(), 'min', json_df)


async def test_unpack_json_max(json_df: DataFrame) -> None:
    await run_test_for(UnpackPolicy.max_number(), 'max', json_df)


async def test_date_component_handle_object_type() -> None:
    input_data = pd.read_csv('test-data/dates.csv')
    result = await Transformations.date_component('day', 'date', 'day_value').transform(input_data)
    assert np.all(result['expected_day'] == result['day_value'])


async def test_date_component_handle_str_type(date_df: DataFrame) -> None:
    date_df['date'] = date_df['date'].astype(str)
    result = await Transformations.date_component('day', 'date', 'day_value').transform(date_df)
    assert np.all(result['expected_day'] == result['day_value'])


async def test_date_component_handle_datetime_type(date_df: DataFrame) -> None:
    date_df['date'] = date_df['date'].apply(lambda date: datetime.fromisoformat(date))
    result = await Transformations.date_component('day', 'date', 'day_value').transform(date_df)
    assert np.all(result['expected_day'] == result['day_value'])


async def test_schema(date_df: DataFrame) -> None:
    schema = {'date': SchemaType.datetime(), 'x_values': SchemaType.integer()}
    result_df = await Transformations.schema(schema).transform(date_df)

    for (key, value) in schema.items():
        assert result_df[key].dtype == value.data_type


async def test_ratio() -> None:
    df = pd.DataFrame(
        {
            'num': [0, np.nan, None, 0, 3, 2],
            'denum': [1, 2, 3, 4, 6, 0],
            'expected': [0, np.nan, np.nan, 0, 0.5, np.nan],
        }
    )
    result_df = await Transformations.ratio('num', 'denum', 'test').transform(df)
    assert result_df['expected'].equals(result_df['test'])



from __future__ import annotations

from typing import Any

from pandas import DataFrame, Series
from sklearn.pipeline import Pipeline


class TrainedModel:

    input_features: list[str]

    def _valid_data(self, data: DataFrame) -> DataFrame:
        feature_set = set(self.input_features)
        intersection = set(data.columns).intersection(feature_set)
        if len(intersection) != len(feature_set):
            raise Exception(f'Missing the following features: {feature_set - intersection}')
        return data[self.input_features]

    def predict(self, data: DataFrame) -> Series:
        return self._predict_on_valid(self._valid_data(data))

    def _predict_on_valid(self, data: DataFrame) -> Series:
        raise NotImplementedError()

    def to_json(self) -> str:
        raise NotImplementedError()

    @staticmethod
    def from_dict(data: dict[str, Any]) -> TrainedModel:
        raise NotImplementedError()


class TrainedEstimatorModel(TrainedModel):
    def estimator(self) -> Pipeline:
        raise NotImplementedError()


class TrainedRegressionModel(TrainedModel):
    """A trained model that preduces regression outputs

    Therefor representing a transformation from
    Vector[float | int] -> float
    """

    pass


class TrainedProbabilisticModel(TrainedModel):
    def predict_proba(self, data: DataFrame) -> DataFrame:
        return self._predict_proba_on_valid(self._valid_data(data))

    def _predict_proba_on_valid(self, data: DataFrame) -> DataFrame:
        raise NotImplementedError()


class TrainedClassificationModel(TrainedModel):
    """A trained model that preduces classification outputs

    Therefor representing a transformation from
    Vector[float | int | bool | str | date | datetime] -> int | bool | str
    """

    def class_names(self) -> list[Any]:
        raise NotImplementedError()


class TrainedProbabilisticClassificationModel(TrainedProbabilisticModel, TrainedClassificationModel):
    def predict_proba(self, data: DataFrame) -> DataFrame:
        label_indicies = self.class_names()
        predictions = self._predict_proba_on_valid(self._valid_data(data))
        return DataFrame(data=predictions, columns=label_indicies)

    def _predict_proba_on_valid(self, data: DataFrame) -> DataFrame:
        raise NotImplementedError()



from aioaws.s3 import S3Config

from bender.model_exporter.aws_s3_exporter import AwsS3ModelExporter
from bender.model_exporter.disk_exporter import DiskExporter


class ModelExporters:
    @staticmethod
    def aws_s3(file_path: str, config: S3Config) -> AwsS3ModelExporter:
        return AwsS3ModelExporter(file_path, config)

    @staticmethod
    def disk(file_path: str) -> DiskExporter:
        return DiskExporter(file_path)



from bender.model_trainer.kneighbors import KNeighborsClassifierTrainer, KNeighborsConfig
from bender.model_trainer.linear_regression import LinearRegressionTrainer
from bender.model_trainer.support_vector import SupportVectorRegressionTrainer, SvrTrainerConfig
from bender.model_trainer.xgboosted_tree import XGBoostTrainer


class Trainers:
    @staticmethod
    def kneighbours(config: KNeighborsConfig = KNeighborsConfig()) -> KNeighborsClassifierTrainer:
        return KNeighborsClassifierTrainer(config)

    @staticmethod
    def xgboost() -> XGBoostTrainer:
        return XGBoostTrainer()

    @staticmethod
    def support_vector_regression(config: SvrTrainerConfig = SvrTrainerConfig.rbf()) -> SupportVectorRegressionTrainer:
        return SupportVectorRegressionTrainer(config)

    @staticmethod
    def linear_regression() -> LinearRegressionTrainer:
        return LinearRegressionTrainer()



import logging

from sklearn.metrics import RocCurveDisplay

from bender.evaluator.interface import Evaluator
from bender.exporter.exporter import Exporter
from bender.split_strategy.split_strategy import TrainingDataSet
from bender.trained_model.interface import TrainedClassificationModel, TrainedEstimatorModel, TrainedModel

logger = logging.getLogger(__name__)


class RocCurve(Evaluator):

    exporter: Exporter

    def __init__(self, exporter: Exporter) -> None:
        self.exporter = exporter

    async def evaluate(self, model: TrainedModel, data_set: TrainingDataSet) -> None:
        if not isinstance(model, TrainedEstimatorModel):
            return
        if isinstance(model, TrainedClassificationModel) and len(model.class_names()) > 2:
            logger.info('RocCurve can not handle multi class classification models')
            return
        display = RocCurveDisplay.from_estimator(
            model.estimator(), data_set.x_validate, data_set.y_validate.astype(float)
        )
        _ = display.ax_.set_title('Roc Curve')
        await self.exporter.store_figure(display.figure_)



from dataclasses import dataclass
from enum import Enum
from typing import Literal, Union

from sklearn.svm import SVR

from bender.model_trainer.interface import ModelTrainer
from bender.split_strategy.split_strategy import TrainingDataSet
from bender.trained_model.interface import TrainedModel
from bender.trained_model.support_vector import SupportVectorRegression


class SvrKernal(Enum):
    LINEAR = 'linear'
    POLY = 'poly'
    RBF = 'rbf'
    SIGMOID = 'sigmoid'
    PRECOMPUTED = 'precomputed'


GammaType = Union[Literal['auto', 'scale'], float]


@dataclass
class SvrTrainerConfig:
    kernal: SvrKernal
    degree: int = 3
    gamma: GammaType = 'scale'
    coef0: float = 0
    tol: float = 0.001
    C: float = 1
    elipson: float = 0.1
    shrinking: bool = True
    cache_size: float = 200
    verbose: bool = False
    max_iter: int = -1

    @staticmethod
    def linear(
        tol: float = 0.001,
        C: float = 1,
        elipson: float = 0.1,
        shrinking: bool = True,
        cache_size: float = 200,
        verbose: bool = False,
        max_iter: int = -1,
    ) -> 'SvrTrainerConfig':
        return SvrTrainerConfig(
            kernal=SvrKernal.LINEAR,
            tol=tol,
            C=C,
            elipson=elipson,
            shrinking=shrinking,
            cache_size=cache_size,
            verbose=verbose,
            max_iter=max_iter,
        )

    @staticmethod
    def poly(
        degree: int = 3,
        gamma: GammaType = 'scale',
        coef0: float = 1,
        tol: float = 0.001,
        C: float = 1,
        elipson: float = 0.1,
        shrinking: bool = True,
        cache_size: float = 200,
        verbose: bool = False,
        max_iter: int = -1,
    ) -> 'SvrTrainerConfig':
        return SvrTrainerConfig(
            kernal=SvrKernal.LINEAR,
            tol=tol,
            C=C,
            elipson=elipson,
            shrinking=shrinking,
            cache_size=cache_size,
            verbose=verbose,
            max_iter=max_iter,
        )

    @staticmethod
    def rbf(
        gamma: GammaType = 'scale',
        tol: float = 0.001,
        C: float = 1,
        elipson: float = 0.1,
        shrinking: bool = True,
        cache_size: float = 200,
        verbose: bool = False,
        max_iter: int = -1,
    ) -> 'SvrTrainerConfig':
        return SvrTrainerConfig(
            gamma=gamma,
            kernal=SvrKernal.LINEAR,
            tol=tol,
            C=C,
            elipson=elipson,
            shrinking=shrinking,
            cache_size=cache_size,
            verbose=verbose,
            max_iter=max_iter,
        )

    @staticmethod
    def sigmoid(
        gamma: GammaType = 'scale',
        coef0: float = 1,
        tol: float = 0.001,
        C: float = 1,
        elipson: float = 0.1,
        shrinking: bool = True,
        cache_size: float = 200,
        verbose: bool = False,
        max_iter: int = -1,
    ) -> 'SvrTrainerConfig':
        return SvrTrainerConfig(
            kernal=SvrKernal.SIGMOID,
            gamma=gamma,
            coef0=coef0,
            tol=tol,
            C=C,
            elipson=elipson,
            shrinking=shrinking,
            cache_size=cache_size,
            verbose=verbose,
            max_iter=max_iter,
        )


class SupportVectorRegressionTrainer(ModelTrainer):

    config: SvrTrainerConfig

    def __init__(self, config: SvrTrainerConfig) -> None:
        self.config = config

    async def train(self, data_split: TrainingDataSet) -> TrainedModel:
        model = SVR(
            kernel=self.config.kernal.value,
            degree=self.config.degree,
            gamma=self.config.gamma,
            coef0=self.config.coef0,
            tol=self.config.tol,
            C=self.config.C,
            epsilon=self.config.elipson,
            shrinking=self.config.shrinking,
            cache_size=self.config.cache_size,
            verbose=self.config.verbose,
            max_iter=self.config.max_iter,
        )
        model.fit(data_split.x_train, data_split.y_train)
        return SupportVectorRegression(model, data_split.x_features)



from typing import Generic, TypeVar

from bender.split_strategy.split_strategy import TrainingDataSet
from bender.trained_model.interface import TrainedModel


class ModelTrainer:
    async def train(self, data_split: TrainingDataSet) -> TrainedModel:
        raise NotImplementedError()


TrainableType = TypeVar('TrainableType')


class Trainable(Generic[TrainableType]):
    def train(self, model: ModelTrainer, input_features: set[str], target_feature: str) -> TrainableType:
        raise NotImplementedError()



from typing import Any, Callable, Optional, Union

from pandas import DataFrame, Series

from bender.transformation.schema import SchemaTransformation, SchemaType
from bender.transformation.transformation import (
    BinaryTransform,
    BinFeature,
    CombineToMean,
    CustomCodeTransformation,
    DateComponent,
    ExponentialShift,
    FillMissingValue,
    FillPolicy,
    Filter,
    LogNormalDistributionShift,
    LogToConsole,
    NeighbourDistance,
    Relation,
    SetIndex,
    SplitString,
    ToCategorical,
    UnpackJson,
    UnpackPolicy,
)


class Transformations:
    @staticmethod
    def log_normal_shift(input: str, output: str, input_has_zeros: bool = True) -> LogNormalDistributionShift:
        return LogNormalDistributionShift(input, output, input_has_zeros)

    @staticmethod
    def date_component(component: str, input: str, output: str) -> DateComponent:
        return DateComponent(component, input, output)

    @staticmethod
    def unpack_json(input: str, key: str, output: str, policy: UnpackPolicy) -> UnpackJson:
        return UnpackJson(input, key, output, policy)

    @staticmethod
    def neighour_distance(
        number_of_neighbours: int,
        latitude: str = 'latitude',
        longitude: str = 'longitude',
        to: Optional[Callable[[DataFrame], Series]] = None,
    ) -> NeighbourDistance:
        return NeighbourDistance(number_of_neighbours, latitude, longitude, to)

    @staticmethod
    def compute(output: str, computation: Callable[[DataFrame], Series]) -> BinaryTransform:
        return BinaryTransform(output, computation)

    @staticmethod
    def fill_missing(feature: str, policy: FillPolicy) -> FillMissingValue:
        return FillMissingValue(feature, policy)

    @staticmethod
    def to_categorical(feature: str, output: Optional[str] = None) -> ToCategorical:
        return ToCategorical(feature, feature if output is None else output)

    @staticmethod
    def log_to_console(data: Callable[[DataFrame], Any]) -> LogToConsole:
        return LogToConsole(data)

    @staticmethod
    def filter(lambda_function: Callable[[DataFrame], Series]) -> Filter:
        return Filter(lambda_function)

    @staticmethod
    def ratio(numerator: str, denumirator: str, output: str) -> Relation:
        return Relation(numerator, denumirator, output)

    @staticmethod
    def exp_shift(value: str, output: str) -> ExponentialShift:
        return ExponentialShift(value, output)

    @staticmethod
    def custom(code: Callable[[DataFrame], DataFrame]) -> CustomCodeTransformation:
        return CustomCodeTransformation(code)

    @staticmethod
    def combined_mean(features: set[str], output: str) -> CombineToMean:
        return CombineToMean(output, list(features))

    @staticmethod
    def bin(feature: str, n_bins: int, output: str) -> BinFeature:
        return BinFeature(feature, output, n_bins)

    @staticmethod
    def set_index(feature: str) -> SetIndex:
        return SetIndex(feature)

    @staticmethod
    def split_string(
        feature: str, seperator: str, output: Union[list[str], str], select_number: int = 1
    ) -> SplitString:
        if isinstance(output, str):
            output = [output]
        return SplitString(feature, output, seperator, select_number)

    @staticmethod
    def schema(schema: dict[str, SchemaType]) -> SchemaTransformation:
        return SchemaTransformation(schema)



from __future__ import annotations

from dataclasses import dataclass
from typing import Generic, Optional, TypeVar

import numpy as np
from pandas import DataFrame, Series


@dataclass
class TrainingDataSet:

    x_features: list[str]
    y_feature: str

    train: DataFrame
    validation: DataFrame

    @property
    def y_train(self) -> Series:
        return self.train[self.y_feature]

    @property
    def y_validate(self) -> Series:
        return self.validation[self.y_feature]

    @property
    def x_train(self) -> DataFrame:
        return self.train[self.x_features]

    @property
    def x_validate(self) -> DataFrame:
        return self.validation[self.x_features]


class SplitStrategy:
    async def split(self, df: DataFrame) -> tuple[DataFrame, DataFrame]:
        raise NotImplementedError()


class UniformSplitRatio(SplitStrategy):

    start_offset: float
    ratio: float
    group_by: str

    def __init__(self, ratio: float, group_by: str, start_offset: float) -> None:
        self.ratio = ratio
        self.group_by = group_by
        self.start_offset = start_offset

    async def split(self, df: DataFrame) -> tuple[DataFrame, DataFrame]:

        train: Optional[DataFrame] = None
        validate: Optional[DataFrame] = None

        for index, group_value in enumerate(df[self.group_by].unique()):
            rows = df.loc[df[self.group_by] == group_value]

            train_ranges: list[tuple[float, float]]
            validate_ranges: list[tuple[float, float]]
            if self.start_offset + self.ratio > 1:
                start = self.start_offset + self.ratio - 1
                train_ranges = [(0, start), (self.start_offset, 1)]
                validate_ranges = [(start, self.start_offset)]
            else:
                train_ranges = [(self.start_offset, self.start_offset + self.ratio)]
                validate_ranges = [(0, self.start_offset), (self.start_offset + self.ratio, 1)]

            for train_range in train_ranges:

                split_start_index = int(round(len(rows) * train_range[0]))
                split_end_index = int(round(len(rows) * train_range[1]))

                if split_end_index == split_start_index:
                    sub_train = rows[split_start_index : split_end_index + 1]
                elif split_start_index == 0:
                    sub_train = rows[:split_end_index]
                elif split_end_index == len(rows):
                    sub_train = rows[split_start_index:]
                else:
                    sub_train = rows[split_start_index:split_end_index]

                if index != 0:
                    train = train.append(sub_train)  # type: ignore
                else:
                    train = sub_train

            for validate_range in validate_ranges:

                split_start_index = int(round(len(rows) * validate_range[0]))
                split_end_index = int(round(len(rows) * validate_range[1]))

                if split_end_index == split_start_index:
                    sub_validate = rows[split_start_index : split_end_index + 1]
                elif split_start_index == 0:
                    sub_validate = rows[:split_end_index]
                elif split_end_index == len(rows):
                    sub_validate = rows[split_start_index:]
                else:
                    sub_validate = rows[split_start_index:split_end_index]

                if index != 0:
                    validate = validate.append(sub_validate)  # type: ignore
                else:
                    validate = sub_validate

        return train, validate


class RandomRatioSplitter(SplitStrategy):

    ratio: float

    def __init__(self, ratio: float) -> None:
        self.ratio = ratio

    async def split(self, df: DataFrame) -> tuple[DataFrame, DataFrame]:
        split_index = int(len(df) * self.ratio)

        train = df[:split_index]
        validate = df[split_index:]

        return train, validate


class SortedRatioSplitter(SplitStrategy):

    sort_key: str
    ratio: float

    def __init__(self, ratio: float, sort_key: str) -> None:
        self.ratio = ratio
        self.sort_key = sort_key

    async def split(self, df: DataFrame) -> tuple[DataFrame, DataFrame]:
        split_index = int(len(df) * self.ratio)
        dates = df[self.sort_key]
        sorted_index = [
            x
            for _, x in sorted(zip(np.array(dates), np.arange(0, len(dates))), key=lambda pair: pair[0])  # type: ignore
        ]

        train_index = sorted_index[:split_index]
        test_index = sorted_index[split_index:]

        train: DataFrame = df.iloc[train_index]
        validate: DataFrame = df.iloc[test_index]

        return train, validate


SplitableType = TypeVar('SplitableType')


class Splitable(Generic[SplitableType]):
    def split(self, split_strategy: SplitStrategy) -> SplitableType:
        raise NotImplementedError()



import pytest
from sklearn.datasets import load_iris

from bender.importers import DataImporters
from bender.metrics import Metrics
from bender.model_exporters import ModelExporters
from bender.model_trainer.xgboosted_tree import XGBoostTrainer
from bender.split_strategies import SplitStrategies

pytestmark = pytest.mark.asyncio


async def test_log_loss_metric() -> None:

    iris = load_iris(as_frame=True)

    data = iris.data
    data['target'] = iris.target

    loss = await (
        DataImporters.literal(data)
        .split(SplitStrategies.ratio(0.7))
        .train(
            XGBoostTrainer(),
            input_features=['sepal length (cm)', 'sepal width (cm)', 'petal length (cm)', 'petal width (cm)'],
            target_feature='target',
        )
        .export_model(ModelExporters.disk('test-exports/test_file.json'))
        .metric(Metrics.log_loss())
        .run()
    )

    assert isinstance(loss, float)
    assert loss > 0



from typing import Generic, TypeVar

from bender.metric.interface import Metric

T = TypeVar('T')


class Metricable(Generic[T]):
    def metric(self, metric: Metric) -> T:
        raise NotImplementedError()



from typing import Generic, TypeVar

from bender.explorer.interface import Explorer

T = TypeVar('T')


class Explorable(Generic[T]):
    def explore(self, explorers: list[Explorer]) -> T:
        raise NotImplementedError()



import matplotlib.pyplot as plt
import seaborn as sns
from pandas import DataFrame

from bender.explorer.interface import Explorer
from bender.exporter.exporter import Exporter


class ViolinPlot(Explorer):

    target: str
    y_feature: str
    exporter: Exporter

    def __init__(self, target: str, y_feature: str, exporter: Exporter) -> None:
        self.exporter = exporter
        self.y_feature = y_feature
        self.target = target

    async def explore(self, df: DataFrame) -> None:
        fig, ax = plt.subplots(figsize=(7, 5))
        sns.violinplot(
            data=df,
            x=self.target,
            y=self.y_feature,
            ax=ax,
        )
        await self.exporter.store_figure(fig)



from __future__ import annotations

from bender.trained_model.interface import TrainedModel


class ModelExporter:
    async def export(self, model: TrainedModel) -> None:
        raise NotImplementedError()



from __future__ import annotations

import json
import logging
from typing import Any, Generic, TypeVar

from aioaws.s3 import S3Client, S3Config
from httpx import AsyncClient

from bender.trained_model.interface import TrainedModel
from bender.trained_model.xgboosted_tree import TrainedXGBoostModel

logger = logging.getLogger(__name__)


class ModelLoader:
    async def load_model(self) -> TrainedModel:
        raise NotImplementedError()

    @staticmethod
    def model_for(name: str) -> type[TrainedModel]:
        models = ModelLoader.supported_models()
        if name not in models:
            raise Exception(f'Unsupported model {name}')
        return models[name]

    @staticmethod
    def supported_models() -> dict[str, type[TrainedModel]]:
        return {'xgboost': TrainedXGBoostModel}


class LiteralLoader(ModelLoader):

    model: TrainedModel

    def __init__(self, model: TrainedModel) -> None:
        self.model = model

    async def load_model(self) -> TrainedModel:
        return self.model


class S3ModelLoader(ModelLoader):

    config: S3Config
    file_path: str

    def __init__(self, file_path: str, config: S3Config) -> None:
        self.file_path = file_path
        self.config = config

    async def load_model(self) -> TrainedModel:
        async with AsyncClient() as client:
            s3_client = S3Client(client, self.config)
            url = s3_client.signed_download_url(path=self.file_path)
            response = await client.get(url)
            logger.info(url)
            response.raise_for_status()
            body = response.content
            body_content: dict[str, Any] = json.loads(body)
            if 'name' not in body_content:
                raise Exception("Unsupported format on model. Missing 'name' parameter")
            model_type = S3ModelLoader.model_for(body_content['name'])
            return model_type.from_dict(body_content)


ModelLoaderType = TypeVar('ModelLoaderType')


class ModelLoadable(Generic[ModelLoaderType]):
    def load_model(self, model: ModelLoader) -> ModelLoaderType:
        raise NotImplementedError()



import logging
from typing import Any, Optional

import matplotlib.pyplot as plt
import seaborn as sns

from bender.evaluator.interface import Evaluator
from bender.exporter.exporter import Exporter
from bender.split_strategy.split_strategy import TrainingDataSet
from bender.trained_model.interface import TrainedModel, TrainedProbabilisticClassificationModel

logger = logging.getLogger(__name__)


class ProbabilityForClassification(Evaluator):

    num_bins: Optional[int]
    classification_of_interest: Optional[Any]
    exporter: Exporter

    def __init__(
        self, exporter: Exporter, classification_of_interest: Optional[Any] = None, num_bins: Optional[int] = None
    ) -> None:
        self.exporter = exporter
        self.classification_of_interest = classification_of_interest
        self.num_bins = num_bins

    async def evaluate(self, model: TrainedModel, data_set: TrainingDataSet) -> None:
        if not isinstance(model, TrainedProbabilisticClassificationModel):
            logger.info(
                'ProbabilityForClassification Evaluator will only work for TrainedProbabilisticClassificationModel'
            )
            return
        pred_result = model.predict_proba(data_set.x_validate)
        true_results = data_set.y_validate.reset_index(drop=True)
        if set(true_results.unique().tolist()) == {0, 1}:
            true_results = true_results.astype(bool)

        labels = true_results.unique()
        # Scores compared to true labels
        fig, ax = plt.subplots(1, 1, figsize=(20, 10))
        n_bins = 50 if self.num_bins is None else self.num_bins
        plt.xlim(0, 1)
        sns.histplot(
            data=pred_result,
            x=self.classification_of_interest,
            hue=true_results,
            multiple='stack',
            bins=n_bins,
            legend=False,
            ax=ax,
        )
        plt.legend(title='True Values', labels=[str(label) for label in labels])
        await self.exporter.store_figure(fig)



from bender.model_exporter.interface import ModelExporter
from bender.trained_model.interface import TrainedModel


class DiskExporter(ModelExporter):

    file_url: str

    def __init__(self, file_url: str) -> None:
        self.file_url = file_url

    async def export(self, model: TrainedModel) -> None:
        json_data = model.to_json()
        with open(self.file_url, 'w') as file:
            file.write(json_data)



from typing import Any

from pandas import DataFrame, Series
from sklearn.neighbors import KNeighborsClassifier

from bender.trained_model.interface import TrainedModel, TrainedProbabilisticClassificationModel


class TrainedKNeighboursClassifier(TrainedProbabilisticClassificationModel):

    model: KNeighborsClassifier
    input_features: list[str]

    def __init__(self, model: KNeighborsClassifier, input_features: list[str]) -> None:
        self.model = model
        self.input_features = input_features

    def class_names(self) -> list[Any]:
        return list(self.model.classes_)

    def _predict_on_valid(self, data: DataFrame) -> Series:
        return self.model.predict(data)

    def _predict_proba_on_valid(self, data: DataFrame) -> DataFrame:
        return self.model.predict_proba(data)

    def to_json(self) -> str:
        # path = f'tmp-xgboost-{uuid4()}.json'
        # self.model.save_model(path)
        # with open(path) as file:
        #     model_json = file.read()
        # return json.dumps({'input_features': self.input_features, 'model': model_json})
        raise NotImplementedError()

    @staticmethod
    def from_dict(data: dict[str, Any]) -> TrainedModel:
        raise NotImplementedError()
        # if 'model' not in data:
        #     raise Exception('Unsupported format')

        # temp_file_path = f'tmp-xgboost-{uuid4()}.json'
        # with open(temp_file_path, 'w') as temp_file:
        #     temp_file.write(data['model'])

        # model = XGBClassifier()
        # model.load_model(temp_file_path)
        # return TrainedXGBoostModel(model, data['input_features'])



from typing import Generic, TypeVar

from bender.data_exporter.interface import DataExporter

T = TypeVar('T')


class Extractable(Generic[T]):
    def extract(self, prediction_as: str, metadata: list[str], exporter: DataExporter) -> T:
        raise NotImplementedError()



from __future__ import annotations

import matplotlib.pyplot as plt
from matplotlib.figure import Figure


class Exporter:
    async def store_figure(self, figure: Figure) -> None:
        raise NotImplementedError()

    # @staticmethod
    # def clearml(logger=Logger) -> ClearmlExporter:
    #     return ClearmlExporter(logger)

    @staticmethod
    def disk(path: str) -> LocalDiskExporter:
        return LocalDiskExporter(path)

    @staticmethod
    def in_memory() -> MemoryExporter:
        return MemoryExporter()


class ChainedExporter(Exporter):

    first: Exporter
    second: Exporter

    def __init__(self, first: Exporter, second: Exporter) -> None:
        self.first = first
        self.second = second

    async def store_figure(self, figure: Figure) -> None:
        await self.first.store_figure(figure)
        await self.second.store_figure(figure)


class LocalDiskExporter(Exporter):

    path: str

    def __init__(self, path: str) -> None:
        self.path = path

    async def store_figure(self, figure: Figure) -> None:
        used_path = self.path + '.png'
        figure.savefig(used_path)


class MemoryExporter(Exporter):
    async def store_figure(self, figure: Figure) -> None:
        plt.show()


# class ClearmlExporter(Exporter):

#     logger: Logger

#     def __init__(self, logger: Logger) -> None:
#         self.logger = logger

#     async def store_data_frame(self, df: DataFrame):
#         raise NotImplementedError()

#     async def store_figure(self, figure: Figure):
#         if isinstance(figure, PltFigure):
#             plt.show(block=False)
#             plt.close()
#         elif isinstance(figure, PlotFigure):
#             self.logger.report_plotly(title="plotly plot", series="plotly", figure=figure)
#         else:
#             raise NotImplementedError()



from bender.metric.log_loss import LogLoss
from bender.metric.mean_square_error import MeanSquareError


class Metrics:
    @staticmethod
    def log_loss() -> LogLoss:
        return LogLoss()

    @staticmethod
    def mean_sqare_error() -> MeanSquareError:
        return MeanSquareError()

    @staticmethod
    def mean_absolute_error() -> MeanSquareError:
        return MeanSquareError()

    @staticmethod
    def mean_absolute_percentage_error() -> MeanSquareError:
        return MeanSquareError()



from typing import Any, Optional

from bender.evaluator.confusion_matrix import ConfusionMatrix
from bender.evaluator.difference_distribution import DifferenceDistribution, DifferenceMetric
from bender.evaluator.feature_importance import XGBoostFeatureImportance
from bender.evaluator.precision_recall import PrecisionRecall
from bender.evaluator.predict_probability import ProbabilityForClassification
from bender.evaluator.roc import RocCurve
from bender.exporter.exporter import Exporter


class Evaluators:
    @staticmethod
    def roc_curve(exporter: Exporter = Exporter.in_memory()) -> RocCurve:
        return RocCurve(exporter)

    @staticmethod
    def confusion_matrix(exporter: Exporter = Exporter.in_memory()) -> ConfusionMatrix:
        return ConfusionMatrix(exporter)

    @staticmethod
    def probability_for(
        classification: Optional[Any] = None, num_bins: Optional[int] = None, exporter: Exporter = Exporter.in_memory()
    ) -> ProbabilityForClassification:
        return ProbabilityForClassification(exporter, classification, num_bins)

    @staticmethod
    def feature_importance(exporter: Exporter = Exporter.in_memory()) -> XGBoostFeatureImportance:
        return XGBoostFeatureImportance(exporter)

    @staticmethod
    def precision_recall(exporter: Exporter = Exporter.in_memory()) -> PrecisionRecall:
        return PrecisionRecall(exporter)

    @staticmethod
    def error_distribution(
        metric: DifferenceMetric, exporter: Exporter = Exporter.in_memory()
    ) -> DifferenceDistribution:
        return DifferenceDistribution(metric, exporter)



from pandas import DataFrame

from bender.data_exporter.interface import DataExporter


class DiskDataExporter(DataExporter):

    path: str

    def __init__(self, path: str) -> None:
        self.path = path

    async def export(self, data_frame: DataFrame) -> None:
        data_frame.to_csv(self.path, index=False)



from bender.split_strategy.split_strategy import TrainingDataSet
from bender.trained_model.interface import TrainedModel


class Metric:
    async def metric(self, model: TrainedModel, data_set: TrainingDataSet) -> float:
        raise NotImplementedError()



from dataclasses import dataclass
from enum import Enum
from typing import Optional

import matplotlib.pyplot as plt
import seaborn as sns
from numpy import ceil, floor, sqrt
from pandas import DataFrame

from bender.explorer.interface import Explorer
from bender.exporter.exporter import Exporter


class HistogramStatistic(Enum):
    COUNT = 'count'
    FREQUENCY = 'frequency'
    PROBABILITY = 'probability'
    PERCENT = 'percent'
    DENSITY = 'density'


class HistogramMultiple(Enum):
    LAYER = 'layer'
    DODGE = 'dodge'
    STACK = 'stack'
    FILL = 'fill'


class UnivariantHistogramElement(Enum):
    BARS = 'bars'
    STEP = 'step'
    POLY = 'poly'


@dataclass
class HistogramConfig:
    n_bins: Optional[int] = None
    statistic: HistogramStatistic = HistogramStatistic.COUNT
    multiple: HistogramMultiple = HistogramMultiple.STACK
    element: UnivariantHistogramElement = UnivariantHistogramElement.BARS


class Histogram(Explorer):

    features: Optional[list[str]]
    target: Optional[str]
    config: HistogramConfig
    exporter: Exporter

    def __init__(
        self, features: Optional[list[str]], target: Optional[str], config: HistogramConfig, exporter: Exporter
    ) -> None:
        self.features = features
        self.exporter = exporter
        self.config = config
        self.target = target

    async def explore(self, df: DataFrame) -> None:
        features = list(df.columns)
        if self.features:
            features = self.features
        if self.target:
            x_axs = int(ceil(sqrt(len(features))))
            y_axs = int(ceil(len(features) / x_axs))
            fig, axs = plt.subplots(x_axs, y_axs, figsize=(7, 5))
            for index, feature in enumerate(features):
                sns.histplot(
                    df,
                    x=feature,
                    hue=self.target,
                    multiple=self.config.multiple.value,
                    bins=self.config.n_bins if self.config.n_bins is not None else 'auto',
                    element=self.config.element.value,
                    stat=self.config.statistic.value,
                    ax=axs[index % x_axs, int(floor(index / x_axs))],
                )
            await self.exporter.store_figure(fig)
        else:
            fig, ax = plt.subplots(figsize=(7, 5))
            df.hist(features, ax=ax)
            await self.exporter.store_figure(fig)



from typing import Optional

import pandas
from pandas import DataFrame

from bender.data_importer.interface import DataImporter


class LocalCsvImporter(DataImporter):

    file: str
    seperator: Optional[str]

    def __init__(self, file: str, seperator: Optional[str]) -> None:
        self.file = file
        self.seperator = seperator

    async def import_data(self) -> DataFrame:
        sep = self.seperator if self.seperator is not None else ','
        return pandas.read_csv(self.file, sep=sep)



from typing import Generic, TypeVar

from bender.model_exporter.interface import ModelExporter

ExportedType = TypeVar('ExportedType')


class ModelExportable(Generic[ExportedType]):
    def export_model(self, exporter: ModelExporter) -> ExportedType:
        raise NotImplementedError()



import matplotlib.pyplot as plt
from xgboost import plot_importance

from bender.evaluator.interface import Evaluator
from bender.exporter.exporter import Exporter
from bender.split_strategy.split_strategy import TrainingDataSet
from bender.trained_model.interface import TrainedModel
from bender.trained_model.xgboosted_tree import TrainedXGBoostModel


class XGBoostFeatureImportance(Evaluator):

    exporter: Exporter

    def __init__(self, exporter: Exporter) -> None:
        self.exporter = exporter

    async def evaluate(self, model: TrainedModel, data_set: TrainingDataSet) -> None:
        if not isinstance(model, TrainedXGBoostModel):
            raise Exception('Only supporting feature importance for XGBoost models')

        fig, ax = plt.subplots(1, 1, figsize=(20, 10))
        plot_importance(model.model, ax=ax)
        await self.exporter.store_figure(fig)



from pandas import DataFrame, Series
from sklearn.svm import SVR

from bender.trained_model.interface import TrainedRegressionModel


class SupportVectorRegression(TrainedRegressionModel):

    model: SVR
    input_features: list[str]

    def __init__(self, model: SVR, input_features: list[str]) -> None:
        self.input_features = input_features
        self.model = model

    def _predict_on_valid(self, data: DataFrame) -> Series:
        return self.model.predict(data)



from bender.split_strategy.split_strategy import TrainingDataSet
from bender.trained_model.interface import TrainedModel


class Evaluator:
    async def evaluate(self, model: TrainedModel, data_set: TrainingDataSet) -> None:
        """Evaluates a model

        Args:
            model ([type]): The model to evaluate
            data_set (DataSplit): The data that can be used to evaluate the model
        """
        raise NotImplementedError()



import pytest
from pandas import DataFrame

from bender.importers import DataImporters
from bender.model_exporters import ModelExporters
from bender.model_loaders import ModelLoaders
from bender.model_trainer.xgboosted_tree import XGBoostTrainer
from bender.split_strategies import SplitStrategies

pytestmark = pytest.mark.asyncio


async def test_export_model_pipeline(date_df: DataFrame) -> None:

    model, data_set = await (
        DataImporters.literal(date_df)
        .split(SplitStrategies.ratio(1))
        .train(XGBoostTrainer(), input_features=['y_values', 'x_values'], target_feature='classification')
        .export_model(ModelExporters.disk('test-exports/test_file.json'))
        .run()
    )

    _, _, result = await (DataImporters.literal(date_df).load_model(ModelLoaders.literal(model)).predict().run())

    assert len(result) == len(date_df)



import logging
from enum import Enum
from typing import Any, Optional

from databases import Database
from pandas import DataFrame

from bender.data_importer.importer import LiteralImporter
from bender.data_importer.local_csv import LocalCsvImporter
from bender.pipeline.factory_states import LoadedData  # type: ignore


class DataSets(Enum):
    IRIS = 'iris'
    DIABETES = 'diabets'
    BREAST_CANCER = 'breast_cancer'
    DIGITS = 'digits'
    WINE = 'wine'
    CALIFORNIA_HOUSING_PRICES = 'california_housing_prices'


logger = logging.getLogger(__name__)


class DataImporters:
    def __combine(data) -> DataFrame:
        df = data.data  # type: ignore
        df['target'] = data.target  # type: ignore
        logger.info(f'Loaded data set with target names: {data.target_names}')  # type: ignore
        return df

    @staticmethod
    def sql_with(database: Database, query: str, values: Optional[dict[str, Any]] = None) -> LoadedData:
        from bender.data_importer.importer import SqlImporter

        return LoadedData(SqlImporter(database, query, values=values, should_disconnect=False), [])

    @staticmethod
    def sql(url: str, query: str, values: Optional[dict[str, Any]] = None) -> LoadedData:
        from bender.data_importer.importer import SqlImporter

        return LoadedData(SqlImporter(Database(url), query, values=values, should_disconnect=True), [])

    @staticmethod
    def literal(df: DataFrame) -> LoadedData:
        return LoadedData(LiteralImporter(df), [])

    @staticmethod
    def csv(file: str, seperator: Optional[str] = None) -> LoadedData:
        return LoadedData(LocalCsvImporter(file, seperator), transformations=[])

    @staticmethod
    def data_set(data_set: DataSets) -> LoadedData:
        df: DataFrame
        if data_set == DataSets.IRIS:
            from sklearn.datasets import load_iris

            df = DataImporters.__combine(load_iris(as_frame=True))
        elif data_set == DataSets.BREAST_CANCER:
            from sklearn.datasets import load_breast_cancer

            df = DataImporters.__combine(load_breast_cancer(as_frame=True))
        elif data_set == DataSets.DIABETES:
            from sklearn.datasets import load_diabetes

            df = DataImporters.__combine(load_diabetes(as_frame=True))
        elif data_set == DataSets.DIGITS:
            from sklearn.datasets import load_digits

            df = DataImporters.__combine(load_digits(as_frame=True))
        elif data_set == DataSets.WINE:
            from sklearn.datasets import load_wine

            df = DataImporters.__combine(load_wine(as_frame=True))
        elif data_set == DataSets.CALIFORNIA_HOUSING_PRICES:
            from sklearn.datasets import fetch_california_housing

            df = DataImporters.__combine(fetch_california_housing(as_frame=True))
        else:
            raise Exception('Unable to find the correct data set')
        return DataImporters.literal(df)



from sklearn.metrics import ConfusionMatrixDisplay

from bender.evaluator.interface import Evaluator
from bender.exporter.exporter import Exporter
from bender.split_strategy.split_strategy import TrainingDataSet
from bender.trained_model.interface import TrainedClassificationModel, TrainedEstimatorModel, TrainedModel


class ConfusionMatrix(Evaluator):

    exporter: Exporter

    def __init__(self, exporter: Exporter) -> None:
        self.exporter = exporter

    async def evaluate(self, model: TrainedModel, data_set: TrainingDataSet) -> None:
        if isinstance(model, TrainedEstimatorModel):
            display = ConfusionMatrixDisplay.from_estimator(
                model.estimator(), data_set.x_validate, data_set.y_validate.astype(float)
            )
        elif isinstance(model, TrainedClassificationModel):
            predictions = model.predict(data_set.x_validate)
            display = ConfusionMatrixDisplay.from_predictions(data_set.y_validate, predictions)
        else:
            return

        _ = display.ax_.set_title('Confusion Matrix')
        await self.exporter.store_figure(display.figure_)



from bender.model_loader.model_loader import LiteralLoader, S3Config, S3ModelLoader
from bender.pipeline.factory_states import LoadedModel  # type: ignore
from bender.trained_model.interface import TrainedModel


class ModelLoaders:
    @staticmethod
    def aws_s3(file: str, config: S3Config) -> LoadedModel:
        return LoadedModel(S3ModelLoader(file, config))

    @staticmethod
    def literal(model: TrainedModel) -> LoadedModel:
        return LoadedModel(LiteralLoader(model))



from typing import Generic, TypeVar

from bender.evaluator.interface import Evaluator

EvaluableType = TypeVar('EvaluableType')


class Evaluable(Generic[EvaluableType]):
    def evaluate(self, evaluators: list[Evaluator]) -> EvaluableType:
        raise NotImplementedError()



import json

import numpy as np
import pytest
from pandas import DataFrame

from bender.importers import DataImporters
from bender.model_exporters import ModelExporters
from bender.model_loaders import ModelLoaders
from bender.model_trainer.xgboosted_tree import XGBoostTrainer
from bender.split_strategies import SplitStrategies
from bender.trained_model.xgboosted_tree import TrainedXGBoostModel

pytestmark = pytest.mark.asyncio


async def test_xgboosted_tree(date_df: DataFrame) -> None:

    model, _ = await (
        DataImporters.literal(date_df)
        .split(SplitStrategies.ratio(1))
        .train(XGBoostTrainer(), input_features=['y_values', 'x_values'], target_feature='classification')
        .export_model(ModelExporters.disk('test-exports/test_file.json'))
        .run()
    )

    json_data = model.to_json()
    json_dict = json.loads(json_data)
    loaded_model = TrainedXGBoostModel.from_dict(json_dict)

    _, _, org_result = await DataImporters.literal(date_df).load_model(ModelLoaders.literal(model)).predict().run()

    _, _, loaded_result = await (
        DataImporters.literal(date_df).load_model(ModelLoaders.literal(loaded_model)).predict().run()
    )

    assert np.all(loaded_result == org_result)


async def test_xgboosted_tree_proba(date_df: DataFrame) -> None:

    model, _ = await (
        DataImporters.literal(date_df)
        .split(SplitStrategies.ratio(1))
        .train(XGBoostTrainer(), input_features=['y_values', 'x_values'], target_feature='classification')
        .run()
    )

    _, _, prob_result = (
        await DataImporters.literal(date_df).load_model(ModelLoaders.literal(model)).predict_proba().run()
    )

    _ = (
        await DataImporters.literal(date_df)
        .load_model(ModelLoaders.literal(model))
        .predict_proba()
        .extract({0: '0-prob', 3: '3-prob'}, metadata=['x_values'])
        .run()
    )

    classifications = {0, 1, 2, 3}
    assert set(prob_result.columns) == classifications
    assert len(prob_result[1]) == len(date_df)

    model, _ = await (
        DataImporters.literal(date_df)
        .split(SplitStrategies.ratio(0.7))
        .train(XGBoostTrainer(), input_features=['y_values', 'x_values'], target_feature='bool_classification')
        .run()
    )

    _, _, prob_result = (
        await DataImporters.literal(date_df).load_model(ModelLoaders.literal(model)).predict_proba().run()
    )
    classifications = {True, False}
    assert set(prob_result.columns) == classifications
    assert len(prob_result[True]) == len(date_df)



from __future__ import annotations

from pandas import DataFrame, Series, Timestamp, to_datetime

from bender.transformation.transformation import Transformation


class SchemaType:
    @property
    def data_type(self) -> type:
        raise NotImplementedError()

    def convert(self, series: Series) -> Series:
        raise NotImplementedError()

    @staticmethod
    def integer() -> IntegerType:
        return IntegerType()

    @staticmethod
    def float() -> FloatType:
        return FloatType()

    @staticmethod
    def string() -> StringType:
        return StringType()

    @staticmethod
    def datetime() -> DatetimeType:
        return DatetimeType()


class IntegerType(SchemaType):
    @property
    def data_type(self) -> type:
        return int

    def convert(self, series: Series) -> Series:
        return series.astype(int)


class StringType(SchemaType):
    @property
    def data_type(self) -> type:
        return str

    def convert(self, series: Series) -> Series:
        return series.astype(str)


class FloatType(SchemaType):
    @property
    def data_type(self) -> type:
        return float

    def convert(self, series: Series) -> Series:
        return series.astype(float)


class DatetimeType(SchemaType):
    @property
    def data_type(self) -> type:
        return type(Timestamp)

    def convert(self, series: Series) -> Series:
        if series.dtype == str:
            return to_datetime(series, infer_datetime_format=True)
        return to_datetime(series.astype(str), infer_datetime_format=True)


class SchemaTransformation(Transformation):  # type: ignore

    schemas: dict[str, SchemaType]

    def __init__(self, schemas: dict[str, SchemaType]) -> None:
        self.schemas = schemas

    async def transform(self, df: DataFrame) -> DataFrame:
        for (key, value) in self.schemas.items():
            if df[key].dtype == value.data_type:
                continue
            df[key] = value.convert(df[key])
        return df



from enum import Enum, EnumMeta
from typing import Any, Callable, Optional, TypeVar

from pandas.core.frame import DataFrame, Series


class MetaEnum(EnumMeta):
    """
    Overwrites __contains__ so we can do `"string" in Enum`.
    """

    def __contains__(cls, item: Any) -> bool:
        try:
            cls(item)
        except ValueError:
            return False
        return True


class BaseEnum(Enum, metaclass=MetaEnum):
    pass


class StrEnum(str, BaseEnum):
    def __str__(self) -> str:
        return self.value  # type: ignore

    @classmethod
    def choices(cls) -> list['StrEnum']:
        return list(cls)


class PredictionOutput(StrEnum):

    PROBABILITY = 'probability'
    CLASSIFICATION = 'classification'


class PredictionExtractor:
    @property
    def output(self) -> tuple[str, PredictionOutput]:
        raise NotImplementedError()

    @property
    def needed_features(self) -> set[str]:
        raise NotImplementedError()

    async def extract(self, output: DataFrame) -> None:
        raise NotImplementedError()


PredictableType = TypeVar('PredictableType')


class Predictable:
    def predict(self, on: Optional[Callable[[DataFrame], Series]] = None) -> PredictableType:
        raise NotImplementedError()


ProbPredictableType = TypeVar('ProbPredictableType')


class ProbabilisticPredictable:
    def predict_proba(self, on: Optional[Callable[[DataFrame], Series]] = None) -> ProbPredictableType:
        raise NotImplementedError()



from pandas import DataFrame, Series
from sklearn.linear_model import LinearRegression

from bender.trained_model.interface import TrainedRegressionModel


class TrainedLinearRegression(TrainedRegressionModel):

    model: LinearRegression
    input_features: list[str]

    def __init__(self, model: LinearRegression, input_features: list[str]) -> None:
        self.model = model
        self.input_features = input_features

    def _predict_on_valid(self, data: DataFrame) -> Series:
        return self.model.predict(data)



from typing import Optional

import seaborn as sns
from pandas import DataFrame

from bender.explorer.interface import Explorer
from bender.exporter.exporter import Exporter


class PairPlot(Explorer):

    features: Optional[list[str]]
    target: str
    exporter: Exporter

    def __init__(self, target: str, exporter: Exporter, features: Optional[list[str]]) -> None:
        self.exporter = exporter
        self.target = target
        self.features = features

    async def explore(self, df: DataFrame) -> None:
        plot = sns.pairplot(df, hue=self.target, vars=self.features)
        await self.exporter.store_figure(plot.fig)



import pytest

from bender.evaluators import Evaluators
from bender.explorers import Explorers
from bender.exporter.exporter import LocalDiskExporter
from bender.importers import DataImporters, DataSets
from bender.metrics import Metrics
from bender.model_trainers import Trainers
from bender.split_strategies import SplitStrategies
from bender.transformations import Transformations

pytestmark = pytest.mark.asyncio


async def test_histogram_explorer() -> None:

    input_features = ['sepal length (cm)', 'sepal width (cm)', 'petal length (cm)', 'petal width (cm)']

    _ = await (
        DataImporters.data_set(DataSets.IRIS)
        .explore([Explorers.histogram(input_features, exporter=LocalDiskExporter('test-exports/explorer'))])
        .run()
    )


async def test_histogram_with_target_explorer() -> None:

    input_features = ['pl exp', 'pw exp', 'petal length (cm)']

    exporter = LocalDiskExporter('test-exports/explorer')

    _ = await (
        DataImporters.data_set(DataSets.IRIS)
        .process(
            [
                Transformations.exp_shift('petal length (cm)', output='pl exp'),
                Transformations.exp_shift('petal width (cm)', output='pw exp'),
            ]
        )
        .explore(
            [
                Explorers.histogram(input_features, target='target', exporter=exporter),
                Explorers.correlation(input_features, exporter),
                Explorers.pair_plot('target', input_features, exporter),
                Explorers.scatter('pl exp', 'pw exp', target='target', exporter=exporter),
            ]
        )
        .split(SplitStrategies.uniform_ratio('target', 0.7))
        .train(Trainers.kneighbours(), input_features=input_features, target_feature='target')
        .evaluate(
            [
                Evaluators.confusion_matrix(exporter),
                Evaluators.precision_recall(exporter),
                Evaluators.roc_curve(exporter),
            ]
        )
        .metric(Metrics.log_loss())
        .run()
    )



from sklearn.linear_model import LinearRegression

from bender.model_trainer.interface import ModelTrainer
from bender.split_strategy.split_strategy import TrainingDataSet
from bender.trained_model.interface import TrainedModel
from bender.trained_model.linear_regression import TrainedLinearRegression


class LinearRegressionTrainer(ModelTrainer):
    async def train(self, data_split: TrainingDataSet) -> TrainedModel:
        lin_reg = LinearRegression()
        lin_reg.fit(data_split.x_train, data_split.y_train)
        return TrainedLinearRegression(lin_reg, data_split.x_features)



from cmath import isnan

import pytest
from pandas import DataFrame

from bender.importers import DataImporters
from bender.metrics import Metrics
from bender.model_trainer.xgboosted_tree import XGBoostTrainer

pytestmark = pytest.mark.asyncio


async def test_cross_validation(date_df: DataFrame) -> None:

    score = await (
        DataImporters.literal(date_df)
        .cross_validate(
            'classification',
            3,
            lambda pipeline: pipeline.train(
                XGBoostTrainer(), input_features=['y_values', 'x_values'], target_feature='bool_classification'
            ).metric(Metrics.log_loss()),
        )
        .run()
    )

    assert score > 0


async def test_cross_validation_nan_metric(date_df: DataFrame) -> None:

    score = await (
        DataImporters.literal(date_df)
        .cross_validate(
            'classification',
            3,
            lambda pipeline: pipeline.train(
                XGBoostTrainer(), input_features=['y_values', 'x_values'], target_feature='bool_classification'
            ),
        )
        .run()
    )

    assert isnan(score)



from typing import Any

import numpy as np
from xgboost import XGBClassifier

from bender.model_trainer.interface import ModelTrainer
from bender.split_strategy.split_strategy import TrainingDataSet
from bender.trained_model.interface import TrainedModel
from bender.trained_model.xgboosted_tree import TrainedXGBoostModel


class XGBoostTrainer(ModelTrainer):

    xgboost_parmas: dict[str, Any]

    def __init__(
        self,
        enable_categorical: bool = False,
        use_label_encoder: bool = False,
        learning_rate: float = 0.01,
        max_depth: int = 5,
        n_estimators: int = 400,
        verbosity: float = 0,
        scale_pos_weight: float = 1.0,
        gamma: float = 0,
        min_child_weight: float = 1,
        colsample_bytree: float = 1,
        reg_lambda: float = 1,
        alpha: float = 0,
    ) -> None:
        self.xgboost_parmas = {
            'enable_categorical': enable_categorical,
            'use_label_encoder': use_label_encoder,
            'learning_rate': learning_rate,
            'max_depth': max_depth,
            'n_estimators': n_estimators,
            'verbosity': verbosity,
            'scale_pos_weight': scale_pos_weight,
            'gamma': gamma,
            'min_child_weight': min_child_weight,
            'colsample_bytree': colsample_bytree,
            'reg_lambda': reg_lambda,
            'alpha': alpha,
        }

    async def train(self, data_split: TrainingDataSet) -> TrainedModel:
        # if data_split.y_train.dtype not in [int, bool, str]:
        #     print(data_split.y_train.dtypes)
        #     raise Exception('Training classification model on continuse values. Maybe you want a regression model?')
        model = XGBClassifier(**self.xgboost_parmas)
        if set(data_split.y_train.unique().tolist()) == {1, 0}:
            pos_data = len(data_split.y_train.loc[data_split.y_train == 1])
            neg_data = data_split.y_train.shape[0] - pos_data
            model.scale_pos_weight = int(np.round(neg_data / pos_data))
        model.fit(data_split.x_train, data_split.y_train, eval_set=[(data_split.x_validate, data_split.y_validate)])
        return TrainedXGBoostModel(model, data_split.x_features)



import seaborn as sns
from matplotlib import pyplot as plt
from pandas import Series
from pandas.core.frame import DataFrame

from bender.evaluator.interface import Evaluator
from bender.exporter.exporter import Exporter
from bender.split_strategy.split_strategy import TrainingDataSet
from bender.trained_model.interface import TrainedModel


class DifferenceMetric:
    def difference(self, predict: Series, true_value: Series) -> Series:
        raise NotImplementedError()


class SqareError(DifferenceMetric):
    def difference(self, predict: Series, true_value: Series) -> Series:
        return (true_value - predict) ** 2


class AbsoluteError(DifferenceMetric):
    def difference(self, predict: Series, true_value: Series) -> Series:
        return (true_value - predict).abs()


class AbsolutePercentageError(DifferenceMetric):
    def difference(self, predict: Series, true_value: Series) -> Series:
        return ((true_value - predict) / true_value).abs()


class DifferenceDistribution(Evaluator):

    diff_metrics: DifferenceMetric
    exporter: Exporter

    def __init__(self, diff_metrics: DifferenceMetric, exporter: Exporter) -> None:
        self.diff_metrics = diff_metrics
        self.exporter = exporter

    async def evaluate(self, model: TrainedModel, data_set: TrainingDataSet) -> None:
        pred = model.predict(data_set.x_validate)
        df = DataFrame(data=[], columns=[])
        df['diff'] = self.diff_metrics.difference(pred, data_set.y_validate)

        fig, ax = plt.subplots()
        sns.histplot(df, x='diff', multiple='stack', ax=ax)
        await self.exporter.store_figure(fig)



import numpy as np
import pytest
from pandas.core.frame import DataFrame

from bender.importers import DataImporters

pytestmark = pytest.mark.asyncio


async def test_append_importer(date_df: DataFrame) -> None:

    index = 3
    first = date_df[:index]
    second = date_df[index:]
    df = await DataImporters.literal(first).append(DataImporters.literal(second)).run()

    assert np.all(df == date_df)



from sklearn.metrics import mean_absolute_error

from bender.metric.interface import Metric
from bender.split_strategy.split_strategy import TrainingDataSet
from bender.trained_model.interface import TrainedClassificationModel, TrainedModel


class MeanAbsoluteError(Metric):
    async def metric(self, model: TrainedModel, data_set: TrainingDataSet) -> float:
        if isinstance(model, TrainedClassificationModel):
            raise Exception('Mean Square error is only for Regression models')
        else:
            result = model.predict(data_set.x_validate)
            return float(mean_absolute_error(data_set.y_validate, result))



import json
from typing import Any
from uuid import uuid4

from pandas import DataFrame, Series
from sklearn.pipeline import Pipeline
from xgboost import XGBClassifier

from bender.trained_model.interface import TrainedEstimatorModel, TrainedModel, TrainedProbabilisticClassificationModel


class TrainedXGBoostModel(TrainedProbabilisticClassificationModel, TrainedEstimatorModel):
    """A Trained XGBoosted tree

    Takes a Vector -> Int
    """

    model: XGBClassifier
    input_features: list[str]

    def __init__(self, model: XGBClassifier, input_features: list[str]) -> None:
        self.model = model
        self.input_features = input_features

    def class_names(self) -> list[Any]:
        return list(self.model.classes_)

    def _predict_on_valid(self, data: DataFrame) -> Series:
        return self.model.predict(data)

    def _predict_proba_on_valid(self, data: DataFrame) -> DataFrame:
        return self.model.predict_proba(data)

    def estimator(self) -> Pipeline:
        return self.model

    def to_json(self) -> str:
        path = f'tmp-xgboost-{uuid4()}.json'
        self.model.save_model(path)
        with open(path) as file:
            model_json = file.read()
        return json.dumps({'input_features': self.input_features, 'model': model_json, 'name': 'xgboost'})

    @staticmethod
    def from_dict(data: dict[str, Any]) -> TrainedModel:

        if 'model' not in data:
            raise Exception('Unsupported format')

        temp_file_path = f'tmp-xgboost-{uuid4()}.json'
        with open(temp_file_path, 'w') as temp_file:
            temp_file.write(data['model'])

        model = XGBClassifier()
        model.load_model(temp_file_path)
        return TrainedXGBoostModel(model, data['input_features'])



from typing import Optional

import seaborn as sns
from pandas import DataFrame

from bender.explorer.interface import Explorer
from bender.exporter.exporter import Exporter


class ScatterChartExplorer(Explorer):

    x_key: str
    y_key: str
    target: Optional[str]
    exporter: Exporter

    def __init__(self, x_key: str, y_key: str, target: Optional[str], exporter: Exporter) -> None:
        self.x_key = x_key
        self.y_key = y_key
        self.target = target
        self.exporter = exporter

    async def explore(self, df: DataFrame) -> None:
        g = sns.jointplot(x=self.x_key, y=self.y_key, data=df, hue=self.target)
        await self.exporter.store_figure(g.fig)



from typing import Optional

from bender.explorer.correlation import CorrelationMatrix
from bender.explorer.histogram import Histogram, HistogramConfig
from bender.explorer.pair import PairPlot
from bender.explorer.scatter import ScatterChartExplorer
from bender.explorer.violin import ViolinPlot
from bender.exporter.exporter import Exporter, MemoryExporter


class Explorers:
    @staticmethod
    def histogram(
        features: Optional[list[str]] = None,
        target: Optional[str] = None,
        config: HistogramConfig = HistogramConfig(),
        exporter: Exporter = MemoryExporter(),
    ) -> Histogram:
        return Histogram(features, target, config, exporter)

    @staticmethod
    def correlation(features: Optional[list[str]] = None, exporter: Exporter = MemoryExporter()) -> CorrelationMatrix:
        return CorrelationMatrix(features, exporter)

    @staticmethod
    def pair_plot(target: str, features: Optional[list[str]] = None, exporter: Exporter = MemoryExporter()) -> PairPlot:
        return PairPlot(target, exporter, features)

    @staticmethod
    def distribution(feature: str, target: str, exporter: Exporter = MemoryExporter()) -> ViolinPlot:
        return ViolinPlot(target, feature, exporter)

    @staticmethod
    def scatter(
        x_feature: str, y_feature: str, target: Optional[str] = None, exporter: Exporter = MemoryExporter()
    ) -> ScatterChartExplorer:
        return ScatterChartExplorer(x_feature, y_feature, target, exporter)



import pytest
from pandas.core.frame import DataFrame

from bender.evaluators import Evaluators
from bender.exporter.exporter import Exporter
from bender.importers import DataImporters
from bender.model_trainer.xgboosted_tree import XGBoostTrainer
from bender.split_strategies import SplitStrategies
from bender.trained_model.xgboosted_tree import TrainedXGBoostModel
from bender.transformations import Transformations

pytestmark = pytest.mark.asyncio


async def test_train_pipeline(date_df: DataFrame) -> None:

    pipeline = (
        DataImporters.literal(date_df)
        .process(
            [
                Transformations.log_normal_shift('y_values', 'y_log', input_has_zeros=False),
                Transformations.neighour_distance(number_of_neighbours=2, latitude='lat', longitude='long'),
                Transformations.date_component('day', 'date', output='day_value'),
                Transformations.date_component('month', 'date', output='month_value'),
                Transformations.date_component('year', 'date', output='year_value'),
            ]
        )
        .split(SplitStrategies.uniform_ratio('classification', 0.5))
        .train(
            XGBoostTrainer(), input_features=['x_values', 'day_value', 'month_value'], target_feature='classification'
        )
        .evaluate(
            [
                Evaluators.roc_curve(Exporter.disk('test-exports/train-evaluation-roc.png')),
                Evaluators.confusion_matrix(Exporter.disk('test-exports/train-evaluation-matrix.png')),
                Evaluators.precision_recall(Exporter.disk('test-exports/train-evaluation-prec-recall.png')),
                Evaluators.probability_for(0, exporter=Exporter.disk('test-exports/classification_probability')),
            ]
        )
    )
    model, _ = await pipeline.run()
    assert isinstance(model, TrainedXGBoostModel)



from pandas import DataFrame, Series
from sklearn.tree import DecisionTreeClassifier

from bender.trained_model.interface import TrainedClassificationModel


class TrainedDecisionTreeClassifier(TrainedClassificationModel):

    model: DecisionTreeClassifier

    def __init__(self, model: DecisionTreeClassifier, input_features: list[str]) -> None:
        self.model = model
        self.input_features = input_features

    def classification_indicies(self) -> list[str]:
        return [str(label) for label in self.model.classes_]

    def _predict_proba_on_valid(self, data: DataFrame) -> DataFrame:
        return self.model.predict_proba(data)

    def _predict_on_valid(self, data: DataFrame) -> Series:
        return self.model.predict(data)

    def to_json(self) -> str:
        # print("Printing tree")
        # object = self.model.tree_
        # methods = [method_name for method_name in dir(object) if callable(getattr(object, method_name))]
        # print(methods)
        # print(object.decision_path(np.array([[0, 1]], dtype=np.float32)))
        raise NotImplementedError()



from dataclasses import asdict, dataclass
from typing import Any, Literal, Optional

from sklearn.neighbors import KNeighborsClassifier

from bender.model_trainer.interface import ModelTrainer
from bender.split_strategy.split_strategy import TrainingDataSet
from bender.trained_model.interface import TrainedModel
from bender.trained_model.kneighbours import TrainedKNeighboursClassifier


@dataclass
class KNeighborsConfig:
    n_neighbors: Optional[int] = None

    weights: Literal['uniform', 'distance'] = 'distance'

    algorithm: Literal['auto', 'ball_tree', 'kd_tree', 'brute'] = 'auto'

    leaf_size: int = 30

    p: int = 2

    metric: str = 'minkowski'
    metric_params: Optional[dict[str, Any]] = None

    n_jobs: Optional[int] = None


class KNeighborsClassifierTrainer(ModelTrainer):

    config: KNeighborsConfig

    def __init__(self, config: KNeighborsConfig) -> None:
        self.config = config

    async def train(self, data_split: TrainingDataSet) -> TrainedModel:
        classifications = data_split.y_train.unique()
        dict_config = asdict(self.config)
        if 'n_neighbors' not in dict_config or dict_config['n_neighbors'] is None:
            dict_config['n_neighbors'] = len(classifications)
        model = KNeighborsClassifier(**dict_config)
        model.fit(data_split.x_train, data_split.y_train)
        return TrainedKNeighboursClassifier(model, data_split.x_features)



from sklearn.tree import DecisionTreeClassifier

from bender.model_trainer.interface import ModelTrainer
from bender.split_strategy.split_strategy import TrainingDataSet
from bender.trained_model.decision_tree import TrainedDecisionTreeClassifier
from bender.trained_model.interface import TrainedModel


class DecisionTreeClassifierTrainer(ModelTrainer):
    async def train(self, data_split: TrainingDataSet) -> TrainedModel:
        model = DecisionTreeClassifier()
        model.fit(data_split.x_train, data_split.y_train)
        return TrainedDecisionTreeClassifier(model, data_split.x_features)



from bender.split_strategy.split_strategy import RandomRatioSplitter, SortedRatioSplitter, UniformSplitRatio


class SplitStrategies:
    @staticmethod
    def ratio(ratio: float) -> RandomRatioSplitter:
        return RandomRatioSplitter(min(max(ratio, 0), 1))

    @staticmethod
    def sorted_ratio(sort_key: str, ratio: float) -> SortedRatioSplitter:
        return SortedRatioSplitter(min(max(ratio, 0), 1), sort_key)

    @staticmethod
    def uniform_ratio(feature: str, ratio: float, start_offset: float = 0) -> UniformSplitRatio:
        return UniformSplitRatio(ratio, feature, start_offset)



from __future__ import annotations

import asyncio
import logging
from datetime import datetime
from typing import Any, Generic, Optional, TypeVar

import pandas
from databases import Database
from pandas import DataFrame

from bender.data_importer.interface import DataImporter

logger = logging.getLogger(__name__)


DataImporterType = TypeVar('DataImporterType')


class DataImportable(Generic[DataImporterType]):
    def import_data(self, importer: DataImporter) -> DataImporterType:
        raise NotImplementedError()


class CachedImporter(DataImporter):

    importer: DataImporter
    path: str
    expiration_date: datetime

    def __init__(self, importer: DataImporter, path: str, expiration_date: datetime) -> None:
        self.importer = importer
        self.path = path
        self.expiration_date = expiration_date

    async def import_data(self) -> DataFrame:
        expration_path = self.path + 'expiration.csv'
        file_path = self.path + '.csv'
        try:
            logger.info('Trying to load csv')
            saved_expiration_date = pandas.read_csv(expration_path)
            if pandas.to_datetime(saved_expiration_date['date'].iloc[0]) < datetime.now():
                logger.info('Refreshing source')
                raise Exception('Out of date cache')
            return pandas.read_csv(file_path)
        except Exception as error:
            logger.info(f'Error loading file, so loading from source: {error}')
            expiration = DataFrame({'date': [self.expiration_date]})
            df = await self.importer.import_data()
            df.to_csv(file_path, index=False)
            expiration.to_csv(expration_path)
            return df


class AppendImporter(DataImporter):

    first_importer: DataImporter
    second_importer: DataImporter
    ignore_index: bool

    def __init__(self, first_importer: DataImporter, second_importer: DataImporter, ignore_index: bool) -> None:
        self.first_importer = first_importer
        self.second_importer = second_importer
        self.ignore_index = ignore_index

    async def import_data(self) -> DataFrame:
        first, second = await asyncio.gather(self.first_importer.import_data(), self.second_importer.import_data())
        return first.append(second, ignore_index=True)


class JoinedImporter(DataImporter):

    first_import: DataImporter
    second_import: DataImporter
    join_key: str

    def __init__(self, first_import: DataImporter, second_import: DataImporter, join_key: str) -> None:
        self.first_import = first_import
        self.second_import = second_import
        self.join_key = join_key

    async def import_data(self) -> DataFrame:
        first_frame, second_frame = await asyncio.gather(
            self.first_import.import_data(), self.second_import.import_data()
        )
        return first_frame.join(second_frame, on=self.join_key, how='inner')


class LiteralImporter(DataImporter):

    df: DataFrame

    def __init__(self, df: DataFrame) -> None:
        self.df = df

    async def import_data(self) -> DataFrame:
        return self.df


class SqlImporter(DataImporter):

    query: str
    values: Optional[dict[str, Any]]
    database: Database
    should_disconnect: bool

    def __init__(
        self, database: Database, query: str, values: Optional[dict[str, Any]], should_disconnect: bool
    ) -> None:
        self.query = query
        self.database = database
        self.values = values
        self.should_disconnect = should_disconnect

    async def import_data(self) -> DataFrame:
        if not self.database.is_connected:
            await self.database.connect()
        records = await self.database.fetch_all(self.query, values=self.values)
        if self.should_disconnect:
            await self.database.disconnect()
        return DataFrame.from_records([dict(record) for record in records])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bender',
 'bender.data_exporter',
 'bender.data_importer',
 'bender.data_importer.tests',
 'bender.evaluator',
 'bender.explorer',
 'bender.explorer.tests',
 'bender.exporter',
 'bender.metric',
 'bender.metric.tests',
 'bender.model_exporter',
 'bender.model_exporter.tests',
 'bender.model_loader',
 'bender.model_trainer',
 'bender.model_trainer.tests',
 'bender.pipeline',
 'bender.pipeline.tests',
 'bender.split_strategy',
 'bender.split_strategy.tests',
 'bender.tests',
 'bender.trained_model',
 'bender.trained_model.tests',
 'bender.transformation',
 'bender.transformation.tests']

package_data = \
{'': ['*']}

install_requires = \
['aioaws>=0.12,<0.13',
 'asyncpg>=0.24.0,<0.25.0',
 'databases>=0.5.3,<0.6.0',
 'matplotlib>=3.4.3,<4.0.0',
 'pandas>=1.3.4,<2.0.0',
 'seaborn>=0.11.2,<0.12.0',
 'sklearn>=0.0,<0.1',
 'xgboost>=1.5.0,<2.0.0']

setup_kwargs = {
    'name': 'benderml',
    'version': '0.1.15',
    'description': 'A Python package that makes ML processes easier, faster and less error prone',
    'long_description': '# Bender 🤖\n\nA Python package for faster, safer, and simpler ML processes.\n\n## Installation\n\n`pip install benderml`\n\n**Usage:**\n\n```python\nfrom bender.importers import DataImporters\n\npre_processed_data = await DataImporters.csv("file/to/data.csv").process([...]).run()\n```\n\n## Why use `bender`?\n\nBender will make your machine learning processes, faster, safer, simpler while at the same time making it easy and flexible. This is done by providing a set base component, around the core processes that will take place in a ML pipeline process. While also helping you with type hints about what your next move could be.\n\n## Pipeline Safety\n\nThe whole pipeline is build using generics from Python\'s typing system. Resulting in an improved developer experience, as the compiler can know if your pipeline\'s logic makes sense before it has started.\n\n## Load a data set\n\nBender makes most of the `sklearn` datasets available through the `DataImporters.data_set(...)` importer. Here will you need to pass an enum to define which dataset you want. It is also possible to load the data from sql, append different data sources and cache, and it is as simple as:\n```python\nfrom bender.importers import DataImporters\n\n# Predifined data set\nDataImporters.data_set(DataSets.IRIS)\n\n# Load SQL\nDataImporters.sql("url", "SELECT ...")\n\n# Cache a sql import\nDataImporters.sql("url", "SELECT ...")\n    .cached("path/to/cache")\n    .append(\n        # Add more data from a different source (with same features)\n        DataImporters.sql(...)\n    )\n```\n\n## Processing\nWhen the data has been loaded is usually the next set to process the data in some way. `bender` will therefore provide different components that transforms features. Therefore making it easier to keep your logic consistent over multiple projects.\n\n```python\nfrom bender.transformations import Transformations\n\nDataImporters.data_set(DataSets.IRIS)\n    .process([\n        # pl exp = e^(petal length)\n        Transformations.exp_shift(\'petal length (cm)\', output=\'pl exp\'),\n\n        # Alternative to `exp_shift`\n        Transformations.compute(\'pl exp\', lambda df: np.exp(df[\'petal length (cm)\'])),\n\n        # purchases = mean value of the json price values\n        Transformations.unpack_json("purchases", key="price", output_feature="price", policy=UnpackPolicy.median_number()),\n\n        ...\n    ])\n```\n\n## EDA\n\nFor view how the data is distribuated, is it also possible to explore the data.\n\n```python\nfrom bender.explorers import Explorers\n\nawait (DataImporters.data_set(DataSets.IRIS)\n    .process([...])\n    .explore([\n        # Display all features in a hist\n        Explorers.histogram(target=\'target\'),\n\n        # Display corr matrix and logs which features you could remove\n        Explorers.correlation(input_features),\n\n        # View how features relate in 2D\n        Explorers.pair_plot(\'target\'),\n    ])\n```\n\n## Splitting into train and test sets\nThere are many ways we can train and test, it is therefore easy to choose and switch between how it is done with `bender`.\n\n```python\nfrom bender.split_strategies import SplitStrategies\n\nawait (DataImporters.data_set(DataSets.IRIS)\n    .process([...])\n\n    # Have 70% as train and 30 as test\n    .split(SplitStrategies.ratio(0.7))\n\n    # Have 70% of each target group in train and the rest in test\n    .split(SplitStrategies.uniform_ratio("target", 0.7))\n\n    # Sorts by the key and taks the first 70% as train\n    .split(SplitStrategies.sorted_ratio("target", 0.7))\n```\n\n## Training\nAfter you have split the data set into train and test, then you can train with the following.\n\n```python\nfrom bender.model_trainers import Trainers\n\nawait (DataImporters.data_set(DataSets.IRIS)\n    .split(...)\n    .train(\n        # train kneighbours on the train test\n        Trainers.kneighbours(),\n        input_features=[...],\n        target_feature="target"\n    )\n```\n\n## Evaluate\nAfter you have a model will it be smart to test how well it works.\n\n```python\nfrom bender.evaluators import Evaluators\n\nawait (DataImporters.data_set(DataSets.IRIS)\n    .split(...)\n    .train(...)\n    .evaluate([\n        # Only present the confusion matrix\n        Evaluators.confusion_matrix(),\n        Evaluators.roc_curve(),\n        Evaluators.precision_recall(),\n    ])\n```\n\n## Save model\nAt last would you need to store the model. You can therefore select one of manny exporters.\n```python\nfrom bender.exporters import Exporters\n\nawait (DataImporters.data_set(DataSets.IRIS)\n    .split(...)\n    .train(...)\n    .export_model(Exporters.aws_s3(...))\n```\n\n## Predict\n```python\nModelLoaders\n    .aws_s3("path/to/model", s3_config)\n    .import_data(\n        DataImporters.sql(sql_url, sql_query)\n    )\n    .predict()\n```\n\n## Extract result\n```python\nModelLoaders\n    .aws_s3(...)\n    .import_data(...)\n    .predict()\n    .extract(prediction_as="target", metadata=[\'entry_id\'], exporter=DataExporters.disk("predictions.csv"))\n```\n\n## Examples\nAn example of the IRIS data set which trains a model to perfection\n\n```python\nawait (DataImporters.data_set(DataSets.IRIS)\n    .process([\n        Transformations.exp_shift(\'petal length (cm)\', output=\'pl exp\'),\n        Transformations.exp_shift(\'petal width (cm)\', output=\'pw exp\'),\n    ])\n    .explore([\n        Explorers.histogram(target=\'target\'),\n        Explorers.correlation(input_features),\n        Explorers.pair_plot(\'target\'),\n    ])\n    .split(SplitStrategies.uniform_ratio("target", 0.7))\n    .train(Trainers.kneighbours(), input_features=input_features, target_feature="target")\n    .evaluate([\n        Evaluators.confusion_matrix()\n    ])\n    .metric(Metrics.log_loss())\n    .run())\n```\n\n## XGBoost Example\nBelow is a simple example for training a XGBoosted tree\n```python\nDataImporters.sql(sql_url, sql_query)\n\n    .process([ # Preproces the data\n        # Extract advanced information from json data\n        Transformations.unpack_json("purchases", key="price", output_feature="price", policy=UnpackPolicy.median_number())\n\n        Transformations.log_normal_shift("y_values", "y_log"),\n\n        # Get date values from a date feature\n        Transformations.date_component("month", "date", output_feature="month_value"),\n    ])\n    .split(SplitStrategies.ratio(0.7))\n\n    # Train a XGBoosted Tree model\n    .train(\n        Trainers.xgboost(),\n        input_features=[\'y_log\', \'price\', \'month_value\', \'country\', ...],\n        target_feature=\'did_buy_product_x\'\n    )\n    .evaluate([\n        Evaluators.roc_curve(),\n        Evaluators.confusion_matrix(),\n        Evaluators.precision_recall(\n            # Overwrite where to export the evaluated result\n            Exporter.disk("precision-recall.png")\n        ),\n    ])\n```\n\n## Predicting Example\n\nBelow will a model be loaded from a AWS S3 bucket, preprocess the data, and predict the output.\nThis will also make sure that the features are valid before predicting.\n\n```python\nModelLoaders\n    # Fetch Model\n    .aws_s3("path/to/model", s3_config)\n\n    # Load data\n    .import_data(\n        DataImporters.sql(sql_url, sql_query)\n            # Caching import localy for 1 day\n            .cached("cache/path")\n    )\n    # Preproces the data\n    .process([\n        Transformations.unpack_json(...),\n        ...\n    ])\n    # Predict the values\n    .predict()\n```\n',
    'author': 'Mats E. Mollestad',
    'author_email': 'mats@mollestad.no',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/otovo/bender',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9.7,<4.0.0',
}


setup(**setup_kwargs)



from pandas import DataFrame

from bender.data_importer.interface import DataImporter
from bender.transformation.transformation import Transformation


class DataExplorer:

    data: DataFrame
    importer: DataImporter

    def __init__(self, importer: DataImporter) -> None:
        self.importer = importer
        self.data = DataFrame()

    def add(self, transformation: Transformation) -> None:
        raise NotImplementedError()

    async def import_with(self, importer: DataImporter) -> None:
        raise NotImplementedError()



import logging
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from pandas import DataFrame

from bender.explorer.interface import Explorer
from bender.exporter.exporter import Exporter

logger = logging.getLogger(__name__)


class CorrelationMatrix(Explorer):

    exporter: Exporter
    features: Optional[list[str]]

    def __init__(self, features: Optional[list[str]], exporter: Exporter) -> None:
        self.exporter = exporter
        self.features = features

    async def explore(self, df: DataFrame) -> None:
        if self.features:
            corr_heatmap = df[self.features].corr()
        else:
            corr_heatmap = df.corr()
        corr_threshold = 0.9
        for feature in corr_heatmap.columns:
            is_feature_mask = corr_heatmap.columns == feature
            column_values = corr_heatmap[corr_heatmap.columns == feature]
            heatmap_mask = ((column_values > corr_threshold) | (column_values < -corr_threshold)) & (~is_feature_mask)
            mask = heatmap_mask.iloc[0]
            correlated_featrues = corr_heatmap.columns[mask]
            if len(correlated_featrues) == 0:
                continue
            new_mask = column_values.columns.isin(correlated_featrues)
            logger.info('Warning: Correlated features should be considered to be removed')
            logger.info(
                f'{feature} is related to {correlated_featrues}, corr values: {column_values.iloc[0][new_mask]}'
            )

        fig, ax = plt.subplots(figsize=(10, 8))
        sns.heatmap(
            corr_heatmap,
            mask=np.zeros_like(corr_heatmap, dtype=np.bool),
            cmap=sns.diverging_palette(220, 10, as_cmap=True),
            square=True,
            ax=ax,
        )
        ax.set_title('Correlation Matrix')
        await self.exporter.store_figure(fig)



import pytest

from bender.evaluator.difference_distribution import AbsolutePercentageError, DifferenceDistribution
from bender.explorer.histogram import HistogramConfig, HistogramMultiple
from bender.explorers import Explorers
from bender.exporter.exporter import Exporter
from bender.importers import DataImporters, DataSets
from bender.metrics import Metrics
from bender.model_trainers import Trainers
from bender.split_strategies import SplitStrategies
from bender.transformations import Transformations

pytestmark = pytest.mark.asyncio


async def test_train_regression_model() -> None:

    loss = await (
        DataImporters.data_set(DataSets.CALIFORNIA_HOUSING_PRICES)
        .process([Transformations.bin('target', 4, 'target_bin')])
        .explore(
            [
                Explorers.histogram(
                    target='target_bin',
                    config=HistogramConfig(n_bins=25, multiple=HistogramMultiple.FILL),
                    exporter=Exporter.disk('test-exports/hist-explor'),
                )
            ]
        )
        .split(SplitStrategies.uniform_ratio('target_bin', 0.7))
        .train(Trainers.linear_regression(), input_features=['MedInc', 'Population'], target_feature='target')
        .evaluate([DifferenceDistribution(AbsolutePercentageError(), Exporter.disk('test-exports/diff-dist'))])
        .metric(Metrics.mean_absolute_percentage_error())
        .run()
    )
    assert loss > 0



from aioaws.s3 import S3Client, S3Config
from httpx import AsyncClient

from bender.model_exporter.interface import ModelExporter
from bender.trained_model.interface import TrainedModel


class AwsS3ModelExporter(ModelExporter):

    config: S3Client
    file_path: str

    def __init__(self, file_path: str, config: S3Config) -> None:
        self.file_path = file_path
        self.config = config

    async def export(self, model: TrainedModel) -> None:
        async with AsyncClient() as client:
            s3_clinet = S3Client(client, self.config)
            file_content = model.to_json()
            byte_data = bytes(file_content, 'utf-8')
            await s3_clinet.upload(self.file_path, byte_data)



from __future__ import annotations

import asyncio

from pandas import DataFrame


class DataExporter:
    async def export(self, data_frame: DataFrame) -> None:
        raise NotImplementedError()

    def and_export(self, exporter: DataExporter) -> ChainedExporter:
        return ChainedExporter(self, exporter)


class ChainedExporter(DataExporter):

    first_exporter: DataExporter
    second_exporter: DataExporter

    def __init__(self, first: DataExporter, second: DataExporter) -> None:
        self.first_exporter = first
        self.second_exporter = second

    async def export(self, data_frame: DataFrame) -> None:
        await asyncio.gather(self.first_exporter.export(data_frame), self.second_exporter.export(data_frame))



import numpy as np
import pytest
from pandas.core.frame import DataFrame

from bender.importers import DataImporters
from bender.model_loaders import ModelLoaders
from bender.model_trainer.decision_tree import DecisionTreeClassifierTrainer
from bender.split_strategies import SplitStrategies

pytestmark = pytest.mark.asyncio


async def test_predict_data() -> None:

    model, data_set = await (
        DataImporters.literal(DataFrame({'x': [0, 1], 'y': [0, 1], 'output': [0, 1]}))
        # No test set
        .split(SplitStrategies.ratio(1))
        .train(DecisionTreeClassifierTrainer(), input_features=['x', 'y'], target_feature='output')
        .run()
    )

    test_data = DataFrame({'x': [2, -3, 4], 'y': [2, -3, 4]})
    expected = [1, 0, 1]
    _, _, result = await (ModelLoaders.literal(model).import_data(DataImporters.literal(test_data)).predict().run())

    assert np.all(expected == result)

    """
    Supervised Regression

    Vector[float] -> float

    .train(
        RegresionModels.linear(),
        input_features=["area", "location"], # floats
        target_feature="price" # float
    )
    """

    """
    Supervised Classification

    Vector[float / int / bool / str] -> str / bool / int

    .train(
        ClassificationModels.DecisionTree(),
        input_features=["sepal_length", "sepal_width"], # float / int / bool / str
        target_feature="class_name" # str / bool / int
    )

    # Should only be avaialbe for clustering / classification problems
    .predict_probability(
        labels={
            "setosa": "is_setosa_probability",
            "versicolor": "is_versicolor_probability",
        }
    )
    """
